"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

model_data.py
~~~~~~~~~~~~~~~~~~

Functionality to build the model-internal data array and process
time-varying parameters.

"""

import collections

import ruamel.yaml
import xarray as xr
import numpy as np
import pandas as pd

from calliope.core.attrdict import AttrDict
from calliope._version import __version__
from calliope.preprocess import checks
from calliope.preprocess.util import split_loc_techs_transmission, concat_iterable
from calliope.preprocess.time import add_time_dimension
from calliope.preprocess.lookup import add_lookup_arrays


def build_model_data(model_run, debug=False):
    """
    Take a Calliope model_run and convert it into an xarray Dataset, ready for
    constraint generation. Timeseries data is also extracted from file at this
    point, and the time dimension added to the data

    Parameters
    ----------
    model_run : AttrDict
        preprocessed model_run dictionary, as produced by
        Calliope.preprocess.preprocess_model
    debug : bool, default = False
        Used to debug steps within build_model_data, particularly before/after
        time dimension addition. If True, more information is returned

    Returns
    -------
    data : xarray Dataset
        Dataset with optimisation parameters as variables, optimisation sets as
        coordinates, and other information in attributes.
    data_dict : dict, only returned if debug = True
        dictionary of parameters, prior to time dimension addition. Used here to
        populate the Dataset (using `from_dict()`)
    data_pre_time : xarray Dataset, only returned if debug = True
        Dataset, prior to time dimension addition, with optimisation parameters
        as variables, optimisation sets as coordinates, and other information
        in attributes.
    """
    # We build up a dictionary of the data, then convert it to an xarray Dataset
    # before applying time dimensions
    data = xr.Dataset(coords=add_sets(model_run), attrs=add_attributes(model_run))

    data_dict = dict()
    data_dict.update(constraints_to_dataset(model_run))
    data_dict.update(costs_to_dataset(model_run))
    data_dict.update(location_specific_to_dataset(model_run))
    data_dict.update(tech_specific_to_dataset(model_run))
    data_dict.update(carrier_specific_to_dataset(model_run))
    data_dict.update(group_constraints_to_dataset(model_run))

    data = data.merge(xr.Dataset.from_dict(data_dict))

    data = add_lookup_arrays(data, model_run)

    if debug:
        data_pre_time = data.copy(deep=True)

    data = add_time_dimension(data, model_run)

    data = update_dtypes(data)
    # Carrier information uses DataArray indexing in the function, so we merge
    # these directly into the main xarray Dataset

    if debug:
        return data, data_dict, data_pre_time
    else:
        return data


def add_sets(model_run):
    coords = dict()
    for key, value in model_run.sets.items():
        if value:
            coords[key] = value
    for key, value in model_run.constraint_sets.items():
        if value:
            coords[key] = value
    return coords


def constraints_to_dataset(model_run):
    """
    Extract all constraints from the processed dictionary (model.model_run) and
    return an xarray Dataset with all the constraints as DataArray variables and
    model sets as Dataset dimensions.

    Parameters
    ----------
    model_run : AttrDict
        processed Calliope model_run dict

    Returns
    -------
    data_dict : dict conforming to xarray conventions

    """
    data_dict = dict()

    # FIXME: hardcoding == bad
    def _get_set(constraint):
        """
        return the set of loc_techs over which the given constraint should be
        built
        """
        if "_area" in constraint:
            return "loc_techs_area"
        elif any(
            i in constraint for i in ["resource_cap", "parasitic", "resource_min_use"]
        ):
            return "loc_techs_supply_plus"
        elif (
            "resource" in constraint
        ):  # i.e. everything with 'resource' in the name that isn't resource_cap
            return "loc_techs_finite_resource"
        elif (
            "storage" in constraint
            or "charge_rate" in constraint
            or "energy_cap_per_storage_cap" in constraint
        ):
            return "loc_techs_store"
        elif "purchase" in constraint:
            return "loc_techs_purchase"
        elif "units_" in constraint:
            return "loc_techs_milp"
        elif "export" in constraint:
            return "loc_techs_export"
        else:
            return "loc_techs"

    # find all constraints which are actually defined in the yaml file
    relevant_constraints = set(
        i.split(".constraints.")[1]
        for i in model_run.locations.as_dict_flat().keys()
        if ".constraints." in i and ".carrier_ratios." not in i
    )
    for constraint in relevant_constraints:
        data_dict[constraint] = dict(dims=_get_set(constraint), data=[])
        for loc_tech in model_run.sets[_get_set(constraint)]:
            loc, tech = loc_tech.split("::", 1)
            # for transmission technologies, we also need to go into link nesting
            if ":" in tech:  # i.e. transmission technologies
                tech, link = tech.split(":")
                loc_tech_dict = model_run.locations[loc].links[link].techs[tech]
            else:  # all other technologies
                loc_tech_dict = model_run.locations[loc].techs[tech]
            constraint_value = loc_tech_dict.constraints.get(constraint, np.nan)
            # inf is assumed to be string on import, so we need to np.inf it
            if constraint_value == "inf":
                constraint_value = np.inf
            # add the value for the particular location & technology combination to the list
            data_dict[constraint]["data"].append(constraint_value)
        # once we've looped through all technology & location combinations, add the array to the dataset

    group_share_data = {}
    group_constraints = ["energy_cap_min", "energy_cap_max", "energy_cap_equals"]
    group_constraints_carrier = [
        "carrier_prod_min",
        "carrier_prod_max",
        "carrier_prod_equals",
        "carrier_con_min",
        "carrier_con_max",
        "carrier_con_equals",
    ]

    for constraint in [  # Only process constraints that are defined
        c
        for c in group_constraints
        if c
        in "".join(model_run.model.get_key("group_share", AttrDict()).keys_nested())
    ]:
        group_share_data[constraint] = [
            model_run.model.get_key(
                "group_share.{}.{}".format(techlist, constraint), np.nan
            )
            for techlist in model_run.sets["techlists"]
        ]

    for constraint in [  # Only process constraints that are defined
        c
        for c in group_constraints_carrier
        if c
        in "".join(model_run.model.get_key("group_share", AttrDict()).keys_nested())
    ]:
        group_share_data[constraint] = [
            [
                model_run.model.get_key(
                    "group_share.{}.{}.{}".format(techlist, constraint, carrier), np.nan
                )
                for techlist in model_run.sets["techlists"]
            ]
            for carrier in model_run.sets["carriers"]
        ]

    # Add to data_dict and set dims correctly
    for k in group_share_data:
        data_dict["group_share_" + k] = {
            "data": group_share_data[k],
            "dims": "techlists"
            if k in group_constraints
            else ("carriers", "techlists"),
        }

    return data_dict


def costs_to_dataset(model_run):
    """
    Extract all costs from the processed dictionary (model.model_run) and
    return an xarray Dataset with all the costs as DataArray variables. Variable
    names will be prepended with `cost_` to differentiate from other constraints

    Parameters
    ----------
    model_run : AttrDict
        processed Calliope model_run dict

    Returns
    -------
    data_dict : dict conforming to xarray conventions

    """
    data_dict = dict()

    # FIXME: hardcoding == bad
    def _get_set(cost):
        """
        return the set of loc_techs over which the given cost should be built
        """
        if any(i in cost for i in ["_cap", "depreciation_rate", "purchase", "area"]):
            return "loc_techs_investment_cost"
        elif any(i in cost for i in ["om_", "export"]):
            return "loc_techs_om_cost"
        else:
            return "loc_techs"

    # find all cost classes and associated costs which are actually defined in the model_run
    costs = set(
        i.split(".costs.")[1].split(".")[1]
        for i in model_run.locations.as_dict_flat().keys()
        if ".costs." in i
    )
    cost_classes = model_run.sets["costs"]

    # loop over unique costs, cost classes and technology & location combinations
    for cost in costs:
        data_dict["cost_" + cost] = dict(dims=["costs", _get_set(cost)], data=[])
        for cost_class in cost_classes:
            cost_class_array = []
            for loc_tech in model_run.sets[_get_set(cost)]:
                loc, tech = loc_tech.split("::", 1)
                # for transmission technologies, we also need to go into link nesting
                if ":" in tech:  # i.e. transmission technologies
                    tech, link = tech.split(":")
                    loc_tech_dict = model_run.locations[loc].links[link].techs[tech]
                else:  # all other technologies
                    loc_tech_dict = model_run.locations[loc].techs[tech]
                cost_dict = loc_tech_dict.get_key("costs." + cost_class, None)

                # inf is assumed to be string on import, so need to np.inf it
                cost_value = np.nan if not cost_dict else cost_dict.get(cost, np.nan)
                # add the value for the particular location & technology combination to the correct cost class list
                cost_class_array.append(cost_value)
            data_dict["cost_" + cost]["data"].append(cost_class_array)

    return data_dict


def carrier_specific_to_dataset(model_run):
    """
    Extract carrier information from the processed dictionary (model.model_run)
    and return an xarray Dataset with DataArray variables describing carrier_in,
    carrier_out, and carrier_ratio (for conversion plus technologies) information.

    Parameters
    ----------
    model_run : AttrDict
        processed Calliope model_run dict

    Returns
    -------
    data_dict : dict conforming to xarray conventions

    """
    carrier_tiers = model_run.sets["carrier_tiers"]
    loc_tech_dict = {k: [] for k in model_run.sets["loc_techs_conversion_plus"]}
    data_dict = dict()
    # Set information per carrier tier ('out', 'out_2', 'in', etc.)
    # for conversion-plus technologies
    if model_run.sets["loc_techs_conversion_plus"]:
        # carrier ratios are the floating point numbers used to compare one
        # carrier_in/_out value with another carrier_in/_out value
        data_dict["carrier_ratios"] = dict(
            dims=["carrier_tiers", "loc_tech_carriers_conversion_plus"], data=[]
        )
        for carrier_tier in carrier_tiers:
            data = []
            for loc_tech_carrier in model_run.sets["loc_tech_carriers_conversion_plus"]:
                loc, tech, carrier = loc_tech_carrier.split("::")
                carrier_ratio = (
                    model_run.locations[loc]
                    .techs[tech]
                    .constraints.get_key(
                        "carrier_ratios.carrier_" + carrier_tier + "." + carrier, 1
                    )
                )
                data.append(carrier_ratio)
                loc_tech_dict[loc + "::" + tech].append(carrier_ratio)
            data_dict["carrier_ratios"]["data"].append(data)

    # Additional system-wide constraints from model_run.model
    if model_run.model.get("reserve_margin", {}) != {}:
        data_dict["reserve_margin"] = {
            "data": [
                model_run.model.reserve_margin.get(c, np.nan)
                for c in model_run.sets["carriers"]
            ],
            "dims": "carriers",
        }

    return data_dict


def location_specific_to_dataset(model_run):
    """
    Extract location specific information from the processed dictionary
    (model.model_run) and return an xarray Dataset with DataArray variables
    describing distance, coordinate and available area information.

    Parameters
    ----------
    model_run : AttrDict
        processed Calliope model_run dict

    Returns
    -------
    data_dict : dict conforming to xarray conventions

    """
    # for every transmission technology, we extract distance information, if it
    # is available
    data_dict = dict()

    data_dict["distance"] = dict(
        dims="loc_techs_transmission",
        data=[
            model_run.get_key(
                "locations.{loc_from}.links.{loc_to}.techs.{tech}.distance".format(
                    **split_loc_techs_transmission(loc_tech)
                ),
                np.nan,
            )
            for loc_tech in model_run.sets["loc_techs_transmission"]
        ],
    )
    # If there is no distance information stored, distance array is deleted
    if data_dict["distance"]["data"].count(np.nan) == len(
        data_dict["distance"]["data"]
    ):
        del data_dict["distance"]

    data_dict["lookup_remotes"] = dict(
        dims="loc_techs_transmission",
        data=concat_iterable(
            [
                (k["loc_to"], k["tech"], k["loc_from"])
                for k in [
                    split_loc_techs_transmission(loc_tech)
                    for loc_tech in model_run.sets["loc_techs_transmission"]
                ]
            ],
            ["::", ":"],
        ),
    )
    # If there are no remote locations stored, lookup_remotes array is deleted
    if data_dict["lookup_remotes"]["data"].count(np.nan) == len(
        data_dict["lookup_remotes"]["data"]
    ):
        del data_dict["lookup_remotes"]

    data_dict["available_area"] = dict(
        dims="locs",
        data=[
            model_run.locations[loc].get("available_area", np.nan)
            for loc in model_run.sets["locs"]
        ],
    )

    # remove this dictionary element if nothing is defined in it
    if set(data_dict["available_area"]["data"]) == {np.nan}:
        del data_dict["available_area"]

    # Coordinates are defined per location, but may not be defined at all for
    # the model
    if "coordinates" in model_run.sets:
        data_dict["loc_coordinates"] = dict(dims=["locs", "coordinates"], data=[])
        for loc in model_run.sets["locs"]:
            data_dict["loc_coordinates"]["data"].append(
                [
                    model_run.locations[loc].coordinates[coordinate]
                    for coordinate in model_run.sets.coordinates
                ]
            )

    return data_dict


def tech_specific_to_dataset(model_run):
    """
    Extract technology (location inspecific) information from the processed
    dictionary (model.model_run) and return an xarray Dataset with DataArray
    variables describing color and inheritance chain information.

    Parameters
    ----------
    model_run : AttrDict
        processed Calliope model_run dict

    Returns
    -------
    data_dict : dict conforming to xarray conventions

    """
    data_dict = collections.defaultdict(lambda: {"dims": ["techs"], "data": []})

    systemwide_constraints = set(
        [
            k.split(".")[-1]
            for k in model_run.techs.keys_nested()
            if ".constraints." in k and k.endswith("_systemwide")
        ]
    )

    for tech in model_run.sets["techs"]:
        if tech in model_run.sets["techs_transmission"]:
            tech = tech.split(":")[0]
        data_dict["colors"]["data"].append(
            model_run.techs[tech].get_key("essentials.color")
        )
        data_dict["inheritance"]["data"].append(
            ".".join(model_run.techs[tech].get_key("inheritance"))
        )
        data_dict["names"]["data"].append(
            # Default to tech ID if no name is set
            model_run.techs[tech].get_key("essentials.name", tech)
        )
        for k in systemwide_constraints:
            data_dict[k]["data"].append(
                model_run.techs[tech].constraints.get_key(k, np.nan)
            )

    return data_dict


def group_constraints_to_dataset(model_run):
    data_dict = {}

    group_constraints = model_run["group_constraints"]

    for constr_name in model_run.constraint_sets["group_constraints"]:
        dims = ["group_names_" + constr_name]
        constr = checks.DEFAULTS.group_constraints.default_group.get(constr_name, None)
        if constr is not None:
            group_names = model_run.constraint_sets.get(
                f"group_names_{constr_name}", None
            )
            if isinstance(constr, dict):
                if "default_carrier" in constr.keys():
                    data = [
                        list(group_constraints[i][constr_name].as_dict_flat().values())[
                            0
                        ]
                        for i in group_names
                    ]
                elif "default_cost" in constr.keys():
                    dims.append("costs")
                    data = [
                        [
                            group_constraints[i][constr_name].get(cost, np.nan)
                            for cost in model_run.sets["costs"]
                        ]
                        for i in group_names
                    ]
            else:
                data = [group_constraints[i][constr_name] for i in group_names]
        else:
            continue

        data_dict["group_" + constr_name] = {"dims": dims, "data": data}

    return data_dict


def add_attributes(model_run):
    attr_dict = AttrDict()

    attr_dict["calliope_version"] = __version__
    attr_dict["applied_overrides"] = model_run["applied_overrides"]
    attr_dict["scenario"] = model_run["scenario"]

    ##
    # Build the `defaults` attribute that holds all default settings
    # used in get_param() lookups inside the backend
    ##

    default_tech_dict = checks.DEFAULTS.techs.default_tech.as_dict()
    default_location_dict = checks.DEFAULTS.locations.default_location.as_dict()

    # Group constraint defaults are a little bit more involved
    default_group_constraint_keys = [
        i
        for i in checks.DEFAULTS.group_constraints.default_group.keys()
        if i not in ["locs", "techs", "exists"]
    ]
    default_group_constraint_dict = {}
    for k in default_group_constraint_keys:
        k_default = checks.DEFAULTS.group_constraints.default_group[k]
        if isinstance(k_default, dict):
            assert len(k_default.keys()) == 1
            default_group_constraint_dict["group_" + k] = k_default[
                list(k_default.keys())[0]
            ]
        else:
            default_group_constraint_dict["group_" + k] = k_default

    attr_dict["defaults"] = ruamel.yaml.dump(
        {
            **default_tech_dict["constraints"],
            **{
                "cost_{}".format(k): v
                for k, v in default_tech_dict["costs"]["default_cost"].items()
            },
            **default_location_dict,
            **default_group_constraint_dict,
        }
    )

    return attr_dict


def update_dtypes(model_data):
    """
    Update dtypes to not be 'Object', if possible.
    Order of preference is: bool, int, float
    """
    for var_name, var in model_data.data_vars.items():
        if var.dtype.kind in ["O", "U"]:
            no_nans = var.where(var != "nan", drop=True)
            model_data[var_name] = var.where(var != "nan")

            if (no_nans.isin(["True", "False", "0", "1"]) | no_nans.isin([0, 1]) | no_nans.isin([True, False])).all():
                # Turn to bool
                model_data[var_name] = var.isin(["True", "1"]) | var.isin([1]) | var.isin([True])
            else:
                try:
                    model_data[var_name] = var.astype(np.int, copy=False)
                except (ValueError, TypeError):
                    try:
                        model_data[var_name] = var.astype(float, copy=False)
                    except ValueError:
                        continue
    return model_data



import pytest  # pylint: disable=unused-import
import numpy as np
import pyomo.core as po
import pandas as pd
import os
import collections
import logging
from itertools import product

from calliope.backend.pyomo.util import get_param
import calliope.exceptions as exceptions
from calliope.core.attrdict import AttrDict

from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import (
    check_error_or_warning,
    check_variable_exists,
    get_indexed_constraint_body,
)


def check_standard_warning(info, warning):

    if warning == "transmission":
        return check_error_or_warning(
            info,
            "dimension loc_techs_transmission and associated variables distance, "
            "lookup_remotes were empty, so have been deleted",
        )


class TestModel:
    @pytest.mark.skip("Buggy")
    @pytest.mark.serial  # Cannot run in parallel with other tests
    def test_load_constraints_no_order(self):
        temp_file = os.path.join(
            os.path.dirname(__file__),
            "..",
            "backend",
            "pyomo",
            "constraints",
            "temp_constraint_file_for_testing.py",
        )

        # Write an empty file
        with open(temp_file, "w") as f:
            f.write("")

        # Should fail, since the empty .py file is included in the list, but
        # has no attribute 'ORDER'.
        with pytest.raises(AttributeError) as excinfo:
            m = build_model({}, "simple_supply,two_hours,investment_costs")
            m.run(build_only=True)

        # We can't use `with` because reasons related to Windows,
        # so we manually remove the temp file in the end
        os.remove(temp_file)

        assert check_error_or_warning(
            excinfo,
            "module 'calliope.backend.pyomo.constraints.temp_constraint_file_for_testing' "
            "has no attribute 'ORDER'",
        )

    @pytest.mark.parametrize(
        "var, domain",
        (
            ("energy_cap_max", "NonNegativeReals"),
            ("resource", "Reals"),
            ("cost_energy_cap", "Reals"),
            ("force_resource", "Boolean"),
            ("names", "Any"),
        ),
    )
    def test_domains(self, var, domain):
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert getattr(m._backend_model, var).domain.name == domain

    def test_first_timestep(self):
        """
        Pyomo likes to 1-index its Sets, which we now expect.
        This test will fail if they ever decide to move to the more pythonic zero-indexing.
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        timestep_0 = "2005-01-01 00:00"
        assert m._backend_model.timesteps.ord(timestep_0) == 1


class TestChecks:
    @pytest.mark.parametrize("on", (True, False))
    def test_operate_cyclic_storage(self, on):
        """Cannot have cyclic storage in operate mode"""
        if on is True:
            override = {}  # cyclic storage is True by default
            m = build_model(
                override, "simple_supply_and_supply_plus,operate,investment_costs"
            )
            assert m.run_config["cyclic_storage"] is True
        elif on is False:
            override = {"run.cyclic_storage": False}
            m = build_model(
                override, "simple_supply_and_supply_plus,operate,investment_costs"
            )
            assert m.run_config["cyclic_storage"] is False
        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)
        check_warn = check_error_or_warning(
            warning, "Storage cannot be cyclic in operate run mode"
        )
        if on is True:
            assert check_warn
        elif on is True:
            assert not check_warn
        assert (
            AttrDict.from_yaml_string(m._model_data.attrs["run_config"]).cyclic_storage
            is False
        )

    @pytest.mark.parametrize(
        "param", [("energy_eff"), ("resource_eff"), ("parasitic_eff")]
    )
    def test_loading_timeseries_operate_efficiencies(self, param):
        m = build_model(
            {
                "techs.test_supply_plus.constraints."
                + param: "file=supply_plus_resource.csv:1"
            },
            "simple_supply_and_supply_plus,operate,investment_costs",
        )
        assert "timesteps" in m._model_data[param].dims

        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)  # will fail to complete run if there's a problem

    def test_operate_group_demand_share_per_timestep_decision(self):
        """Cannot have group_demand_share_per_timestep_decision in operate mode"""
        m = build_model(
            {},
            "simple_supply,investment_costs,operate,enable_group_demand_share_per_timestep_decision",
        )
        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)
        assert check_error_or_warning(
            warning, "`demand_share_per_timestep_decision` group constraints cannot be"
        )
        assert "group_demand_share_per_timestep_decision" not in m._model_data

    @pytest.mark.parametrize("force", (True, False))
    def test_operate_energy_cap_min_use(self, force):
        """If we depend on a finite energy_cap, we have to error on a user failing to define it"""
        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "force_resource": force,
                    "energy_cap_min_use": 0.1,
                    "resource": "file=supply_plus_resource.csv:1",
                    "energy_cap_equals": np.inf,
                }
            },
            "simple_supply_and_supply_plus,operate,investment_costs",
        )

        with pytest.raises(exceptions.ModelError) as error:
            with pytest.warns(exceptions.ModelWarning):
                m.run(build_only=True)

        assert check_error_or_warning(
            error, ["Operate mode: User must define a finite energy_cap"]
        )

    @pytest.mark.parametrize("force", (True, False))
    def test_operate_energy_cap_resource_unit(self, force):
        """If we depend on a finite energy_cap, we have to error on a user failing to define it"""
        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "force_resource": force,
                    "resource_unit": "energy_per_cap",
                    "resource": "file=supply_plus_resource.csv:1",
                    "energy_cap_equals": np.inf,
                    "energy_cap_max": np.inf,
                }
            },
            "simple_supply_and_supply_plus,operate,investment_costs",
        )

        if force is True:
            with pytest.raises(exceptions.ModelError) as error:
                with pytest.warns(exceptions.ModelWarning) as warning:
                    m.run(build_only=True)
            assert check_error_or_warning(
                error, ["Operate mode: User must define a finite energy_cap"]
            )
        elif force is False:
            with pytest.warns(exceptions.ModelWarning) as warning:
                m.run(build_only=True)

    @pytest.mark.parametrize(
        "resource_unit,force",
        list(product(("energy", "energy_per_cap", "energy_per_area"), (True, False))),
    )
    def test_operate_resource_unit_with_resource_area(self, resource_unit, force):
        """Different resource unit affects the capacities which are set to infinite"""
        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "resource_unit": resource_unit,
                    "resource_area_max": 10,
                    "energy_cap_max": 15,
                    "resource": "file=supply_plus_resource.csv:1",
                    "force_resource": force,
                }
            },
            "simple_supply_and_supply_plus,operate,investment_costs",
        )
        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)

        if resource_unit == "energy":
            _warnings = [
                "Energy capacity constraint removed from 0::test_supply_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
                "Resource area constraint removed from 0::test_supply_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
            ]
        elif resource_unit == "energy_per_area":
            _warnings = [
                "Energy capacity constraint removed from 0::test_supply_elec as force_resource is applied and resource is linked to energy flow using `energy_per_area`"
            ]
        elif resource_unit == "energy_per_cap":
            _warnings = [
                "Resource area constraint removed from 0::test_supply_elec as force_resource is applied and resource is linked to energy flow using `energy_per_cap`"
            ]

        if force is True:
            assert check_error_or_warning(warning, _warnings)
        elif force is False:
            assert ~check_error_or_warning(warning, _warnings)

    @pytest.mark.parametrize(
        "resource_unit", [("energy"), ("energy_per_cap"), ("energy_per_area")]
    )
    def test_operate_resource_unit_without_resource_area(self, resource_unit):
        """Different resource unit affects the capacities which are set to infinite"""
        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "resource_unit": resource_unit,
                    "force_resource": True,
                    "resource": "file=supply_plus_resource.csv:1",
                    "energy_cap_max": 15,
                }
            },
            "simple_supply_and_supply_plus,operate,investment_costs",
        )

        with pytest.warns(exceptions.ModelWarning) as warning:
            # energy_per_area without a resource_cap will cause an error, which we have to catch here
            if resource_unit == "energy_per_area":
                with pytest.raises(exceptions.ModelError) as error:
                    m.run(build_only=True)
            else:
                m.run(build_only=True)

        if resource_unit == "energy":
            _warnings = [
                "Energy capacity constraint removed from 0::test_supply_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)"
            ]
            not_warnings = [
                "Resource area constraint removed from 0::test_supply_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
                "Energy capacity constraint removed from 0::test_demand_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
                "Energy capacity constraint removed from 1::test_demand_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
            ]
        elif resource_unit == "energy_per_area":
            _warnings = [
                "Energy capacity constraint removed from 0::test_supply_elec as force_resource is applied and resource is linked to energy flow using `energy_per_area`"
            ]
            not_warnings = [
                "Resource area constraint removed from 0::test_supply_elec as force_resource is applied and resource is linked to energy flow using `energy_per_cap`",
                "Energy capacity constraint removed from 0::test_demand_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
                "Energy capacity constraint removed from 1::test_demand_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
            ]
            # energy_per_area without a resource_cap will cause an error
            check_error_or_warning(
                error,
                "Operate mode: User must define a finite resource_area "
                "(via resource_area_equals or resource_area_max) for 0::test_supply_elec",
            )
        elif resource_unit == "energy_per_cap":
            _warnings = []
            not_warnings = [
                "Resource area constraint removed from 0::test_supply_elec as force_resource is applied and resource is linked to energy flow using `energy_per_cap`",
                "Energy capacity constraint removed from 0::test_supply_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
                "Energy capacity constraint removed from 0::test_demand_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
                "Energy capacity constraint removed from 1::test_demand_elec as force_resource is applied and resource is not linked to energy flow (resource_unit = `energy`)",
            ]
        assert check_error_or_warning(warning, _warnings)
        assert not check_error_or_warning(warning, not_warnings)

    @pytest.mark.parametrize("param", ("charge_rate", "energy_cap_per_storage_cap_max"))
    def test_operate_storage(self, param):
        """Can't violate storage capacity constraints in the definition of a technology"""
        m = build_model(
            {"techs.test_supply_plus.constraints": {param: 0.1}},
            "simple_supply_and_supply_plus,operate,investment_costs",
        )

        with pytest.warns(exceptions.ModelWarning) as warning:
            with pytest.raises(exceptions.ModelError) as error:
                m.run(build_only=True)

        assert check_error_or_warning(
            error,
            "fixed storage capacity * {} is not larger than fixed energy "
            "capacity for loc::tech {}".format(param, "0::test_supply_plus"),
        )
        assert check_error_or_warning(
            warning,
            [
                "Initial stored energy not defined",
                "Resource capacity constraint defined and set to infinity",
                "Storage cannot be cyclic in operate run mode",
            ],
        )

    @pytest.mark.parametrize("on", (True, False))
    def test_operate_resource_cap_max(self, on):
        """Some constraints, if not defined, will throw a warning and possibly change values in model_data"""

        if on is False:
            override = {}
        else:
            override = {"techs.test_supply_plus.constraints.resource_cap_max": 1e6}
        m = build_model(
            override, "simple_supply_and_supply_plus,operate,investment_costs"
        )

        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)
        if on is False:
            assert check_error_or_warning(
                warning, "Resource capacity constraint defined and set to infinity"
            )
            assert np.isinf(
                m._model_data.resource_cap.loc["0::test_supply_plus"].item()
            )
        elif on is True:
            assert not check_error_or_warning(
                warning, "Resource capacity constraint defined and set to infinity"
            )
            assert m._model_data.resource_cap.loc["0::test_supply_plus"].item() == 1e6

    @pytest.mark.parametrize("on", (True, False))
    def test_operate_storage_initial(self, on):
        """Some constraints, if not defined, will throw a warning and possibly change values in model_data"""

        if on is False:
            override = {}
        else:
            override = {"techs.test_supply_plus.constraints.storage_initial": 0.5}
        m = build_model(
            override, "simple_supply_and_supply_plus,operate,investment_costs"
        )

        with pytest.warns(exceptions.ModelWarning) as warning:
            m.run(build_only=True)
        if on is False:
            assert check_error_or_warning(warning, "Initial stored energy not defined")
            assert m._model_data.storage_initial.loc["0::test_supply_plus"].item() == 0
        elif on is True:
            assert not check_error_or_warning(
                warning, "Initial stored energy not defined"
            )
            assert (
                m._model_data.storage_initial.loc["0::test_supply_plus"].item() == 0.5
            )


class TestBalanceConstraints:
    def test_loc_carriers_system_balance_constraint(self):
        """
        sets.loc_carriers
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "system_balance_constraint")

    def test_loc_techs_balance_supply_constraint(self):
        """
        sets.loc_techs_finite_resource_supply,
        """
        m = build_model(
            {"techs.test_supply_elec.constraints.resource": 20},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_supply_constraint")

        m = build_model(
            {
                "techs.test_supply_elec.constraints.resource": 20,
                "techs.test_supply_elec.constraints.resource_unit": "energy_per_cap",
            },
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model, "balance_supply_constraint", "energy_cap"
        )

        m = build_model(
            {
                "techs.test_supply_elec.constraints.resource": 20,
                "techs.test_supply_elec.constraints.resource_unit": "energy_per_area",
            },
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model, "balance_supply_constraint", "resource_area"
        )

    def test_loc_techs_balance_demand_constraint(self):
        """
        sets.loc_techs_finite_resource_demand,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_demand_constraint")

        m = build_model(
            {"techs.test_demand_elec.constraints.resource_unit": "energy_per_cap"},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model, "balance_demand_constraint", "energy_cap"
        )

        m = build_model(
            {"techs.test_demand_elec.constraints.resource_unit": "energy_per_area"},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model, "balance_demand_constraint", "resource_area"
        )

    def test_loc_techs_resource_availability_supply_plus_constraint(self):
        """
        sets.loc_techs_finite_resource_supply_plus,
        """
        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_availability_supply_plus_constraint")

        m = build_model(
            {"techs.test_supply_plus.constraints.resource_unit": "energy_per_cap"},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model,
            "resource_availability_supply_plus_constraint",
            "energy_cap",
        )

        m = build_model(
            {"techs.test_supply_plus.constraints.resource_unit": "energy_per_area"},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model,
            "resource_availability_supply_plus_constraint",
            "resource_area",
        )

    def test_loc_techs_balance_transmission_constraint(self):
        """
        sets.loc_techs_transmission,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_transmission_constraint")

    def test_loc_techs_balance_supply_plus_constraint(self):
        """
        sets.loc_techs_supply_plus,
        """
        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_supply_plus_constraint")

    def test_loc_techs_balance_storage_constraint(self):
        """
        sets.loc_techs_storage,
        """
        m = build_model({}, "simple_storage,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_storage_constraint")
        assert not hasattr(m._backend_model, "storage_initial_constraint")

    def test_loc_techs_balance_storage_discharge_depth_constraint(self):
        """
        sets.loc_techs_storage,
        """
        m = build_model(
            {}, "simple_storage,two_hours,investment_costs,storage_discharge_depth"
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_discharge_depth_constraint")
        assert not hasattr(m._backend_model, "storage_initial_constraint")

        with pytest.raises(exceptions.ModelError) as error:
            m2 = build_model(
                {"techs.test_storage.constraints.storage_initial": 0.0},
                "simple_storage,one_day,investment_costs,storage_discharge_depth",
            )
            m2.run(build_only=True)

        assert check_error_or_warning(
            error, "storage_initial is smaller than storage_discharge_depth."
        )
        m3 = build_model(
            {"techs.test_storage.constraints.storage_initial": 1},
            "simple_storage,one_day,investment_costs,storage_discharge_depth",
        )
        m3.run(build_only=True)
        assert (
            m3._model_data.storage_initial.values
            > m3._model_data.storage_discharge_depth.values
        ).all()

    def test_storage_initial_constraint(self):
        """
        sets.loc_techs_store,
        """
        m = build_model({}, "simple_storage,one_day,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_storage_constraint")
        assert not hasattr(m._backend_model, "storage_initial_constraint")

        m2 = build_model(
            {"techs.test_storage.constraints.storage_initial": 0},
            "simple_storage,one_day,investment_costs",
        )
        m2.run(build_only=True)
        assert hasattr(m2._backend_model, "balance_storage_constraint")
        assert hasattr(m2._backend_model, "storage_initial_constraint")

    def test_carriers_reserve_margin_constraint(self):
        """
        i for i in sets.carriers if i in model_run.model.get_key('reserve_margin', {}).keys()
        """
        m = build_model(
            {"model.reserve_margin.electricity": 0.01},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "reserve_margin_constraint")


class TestCostConstraints:
    # costs.py
    def test_loc_techs_cost_constraint(self):
        """
        sets.loc_techs_cost,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_constraint")

    def test_loc_techs_cost_investment_constraint(self):
        """
        sets.loc_techs_investment_cost,
        """
        m = build_model({}, "simple_conversion,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_investment_constraint")

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_cost_investment_milp_constraint(self):
        m = build_model(
            {
                "techs.test_supply_elec.constraints.lifetime": 10,
                "techs.test_supply_elec.costs.monetary.interest_rate": 0.1,
            },
            "supply_purchase,two_hours",
        )
        m.run(build_only=True)

        assert hasattr(m._backend_model, "cost_investment_constraint")

    def test_loc_techs_not_cost_var_constraint(self):
        """
        i for i in sets.loc_techs_om_cost if i not in sets.loc_techs_conversion_plus + sets.loc_techs_conversion

        """
        m = build_model({}, "simple_conversion,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "cost_var_constraint")

    @pytest.mark.parametrize(
        "tech,scenario,cost",
        (
            ("test_conversion", "simple_conversion", "om_con"),
            ("test_conversion_plus", "simple_conversion_plus", "om_prod"),
        ),
    )
    def test_loc_techs_cost_var_rhs(self, tech, scenario, cost):
        m = build_model(
            {"techs.{}.costs.monetary.{}".format(tech, cost): 1},
            "{},two_hours".format(scenario),
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_rhs")
        assert not hasattr(m._backend_model, "cost_var_constraint")

    @pytest.mark.parametrize(
        "tech,scenario,cost",
        (
            ("test_supply_elec", "simple_supply", "om_prod"),
            ("test_supply_elec", "simple_supply", "om_con"),
            ("test_supply_plus", "simple_supply_and_supply_plus", "om_con"),
            ("test_demand_elec", "simple_supply", "om_con"),
            ("test_transmission_elec", "simple_supply", "om_prod"),
        ),
    )
    def test_loc_techs_cost_var_constraint(self, tech, scenario, cost):
        """
        i for i in sets.loc_techs_om_cost if i not in sets.loc_techs_conversion_plus + sets.loc_techs_conversion

        """
        m = build_model(
            {"techs.{}.costs.monetary.{}".format(tech, cost): 1},
            "{},two_hours".format(scenario),
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_constraint")

    def test_one_way_om_cost(self):
        """
        With one_way transmission, it should still be possible to set an om_prod cost.
        """
        m = build_model(
            {
                "techs.test_transmission_elec.costs.monetary.om_prod": 1,
                "links.0,1.techs.test_transmission_elec.constraints.one_way": True,
            },
            "simple_supply,two_hours",
        )
        m.run(build_only=True)
        arg1 = m._backend_model
        arg2 = "cost_var_constraint"
        arg3 = [
            "monetary",
            "1::test_transmission_elec:0",
            m._backend_model.timesteps[1],
        ]
        has_cost = get_indexed_constraint_body(arg1, arg2, tuple(arg3)).to_string()

        arg3[1] = "0::test_transmission_elec:1"
        has_no_cost = get_indexed_constraint_body(arg1, arg2, tuple(arg3)).to_string()
        assert "cost_om_prod" in has_cost and "carrier_prod" in has_cost
        assert "cost_om_prod" not in has_no_cost and "carrier_prod" not in has_no_cost


class TestExportConstraints:
    # export.py
    def test_loc_carriers_update_system_balance_constraint(self):
        """
        i for i in sets.loc_carriers if sets.loc_techs_export
        and any(['{0}::{2}'.format(*j.split('::')) == i
        for j in sets.loc_tech_carriers_export])
        """

        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        export_exists = check_variable_exists(
            m._backend_model, "system_balance_constraint", "carrier_export"
        )
        assert not export_exists

        m = build_model({}, "supply_export,two_hours,investment_costs")
        m.run(build_only=True)

        export_exists = check_variable_exists(
            m._backend_model, "system_balance_constraint", "carrier_export"
        )
        assert export_exists

    def test_loc_tech_carriers_export_balance_constraint(self):
        """
        sets.loc_tech_carriers_export,
        """

        m = build_model({}, "supply_export,two_hours,investment_costs")
        m.run(build_only=True)

        assert hasattr(m._backend_model, "export_balance_constraint")

    def test_loc_techs_update_costs_var_constraint(self):
        """
        i for i in sets.loc_techs_om_cost if i in sets.loc_techs_export
        """

        m = build_model({}, "supply_export,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_rhs")
        assert hasattr(m._backend_model, "cost_var_constraint")

        m = build_model(
            {"techs.test_supply_elec.costs.monetary.om_prod": 0.1},
            "supply_export,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_rhs")
        assert hasattr(m._backend_model, "cost_var_constraint")

        export_exists = check_variable_exists(
            m._backend_model, "cost_var_constraint", "carrier_export"
        )
        assert export_exists

    def test_loc_tech_carriers_export_max_constraint(self):
        """
        i for i in sets.loc_tech_carriers_export
        if constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.export_cap')
        """

        m = build_model(
            {"techs.test_supply_elec.constraints.export_cap": 5},
            "supply_export,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "export_max_constraint")


class TestCapacityConstraints:
    # capacity.py
    def test_loc_techs_storage_capacity_constraint(self):
        """
        i for i in sets.loc_techs_store if i not in sets.loc_techs_milp
        """
        m = build_model({}, "simple_storage,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_capacity_constraint")

        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_capacity_constraint")

        m = build_model(
            {"techs.test_storage.constraints.storage_cap_equals": 20},
            "simple_storage,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert (
            m._backend_model.storage_capacity_constraint["0::test_storage"].upper()
            == 20
        )
        assert (
            m._backend_model.storage_capacity_constraint["0::test_storage"].lower()
            == 20
        )

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_storage_capacity_milp_constraint(self):
        m = build_model(
            {
                "techs.test_storage.constraints": {
                    "units_max": 1,
                    "energy_cap_per_unit": 20,
                    "storage_cap_per_unit": 20,
                }
            },
            "simple_storage,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "storage_capacity_constraint")

    @pytest.mark.parametrize(
        "scenario,tech,override",
        [
            i + (j,)
            for i in [
                ("simple_supply_and_supply_plus", "test_supply_plus"),
                ("simple_storage", "test_storage"),
            ]
            for j in ["max", "equals", "min"]
        ],
    )
    def test_loc_techs_energy_capacity_storage_constraint(
        self, scenario, tech, override
    ):
        """
        i for i in sets.loc_techs_store if constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_max')
        """
        m = build_model(
            {
                "techs.{}.constraints.energy_cap_per_storage_cap_{}".format(
                    tech, override
                ): 0.5
            },
            "{},two_hours,investment_costs".format(scenario),
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "energy_capacity_storage_{}_constraint".format(override)
        )
        if override == "equals":
            assert not any(
                [
                    hasattr(
                        m._backend_model,
                        "energy_capacity_storage_{}_constraint".format(i),
                    )
                    for i in set(["max", "min"])
                ]
            )

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    @pytest.mark.parametrize("override", (("max", "equals", "min")))
    def test_loc_techs_energy_capacity_milp_storage_constraint(self, override):
        """
        i for i in sets.loc_techs_store if constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_max')
        """

        m = build_model(
            {
                "techs.test_supply_plus.constraints.energy_cap_per_storage_cap_{}".format(
                    override
                ): 0.5
            },
            "supply_and_supply_plus_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "energy_capacity_storage_{}_constraint".format(override)
        )
        if override == "equals":
            assert not any(
                [
                    hasattr(
                        m._backend_model,
                        "energy_capacity_storage_{}_constraint".format(i),
                    )
                    for i in set(["max", "min"])
                ]
            )

    def test_no_loc_techs_energy_capacity_storage_constraint(self, caplog):
        """
        i for i in sets.loc_techs_store if constraint_exists(model_run, i, 'constraints.energy_cap_per_storage_cap_max')
        """
        with caplog.at_level(logging.INFO):
            m = build_model(model_file="energy_cap_per_storage_cap.yaml")

        assert (
            "consider defining a `energy_cap_per_storage_cap_min/max/equals` constraint"
            in caplog.text
        )

        m.run(build_only=True)
        assert not any(
            [
                hasattr(
                    m._backend_model, "energy_capacity_storage_{}_constraint".format(i)
                )
                for i in ["max", "equals", "min"]
            ]
        )

    @pytest.mark.parametrize("override", ((None, "max", "equals", "min")))
    def test_loc_techs_resource_capacity_constraint(self, override):
        """
        i for i in sets.loc_techs_finite_resource_supply_plus
        if any([constraint_exists(model_run, i, 'constraints.resource_cap_equals'),
                constraint_exists(model_run, i, 'constraints.resource_cap_max'),
                constraint_exists(model_run, i, 'constraints.resource_cap_min')])
        """

        if override is None:
            m = build_model(
                {}, "simple_supply_and_supply_plus,two_hours,investment_costs"
            )
            m.run(build_only=True)
            assert not hasattr(m._backend_model, "resource_capacity_constraint")

        else:
            m = build_model(
                {
                    "techs.test_supply_plus.constraints.resource_cap_{}".format(
                        override
                    ): 10
                },
                "simple_supply_and_supply_plus,two_hours,investment_costs",
            )
            m.run(build_only=True)
            assert hasattr(m._backend_model, "resource_capacity_constraint")

    def test_loc_techs_resource_capacity_equals_energy_capacity_constraint(self):
        """
        i for i in sets.loc_techs_finite_resource_supply_plus
        if constraint_exists(model_run, i, 'constraints.resource_cap_equals_energy_cap')
        """

        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "resource_capacity_equals_energy_capacity_constraint"
        )

        m = build_model(
            {"techs.test_supply_plus.constraints.resource_cap_equals_energy_cap": True},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "resource_capacity_equals_energy_capacity_constraint"
        )

    def test_loc_techs_resource_area_constraint(self):
        """
        i for i in sets.loc_techs_area if i in sets.loc_techs_supply_plus
        """

        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "resource_area_constraint")

        m = build_model(
            {"techs.test_supply_plus.constraints.resource_area_max": 10},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_area_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.resource_area_max": 10},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_area_constraint")

        # Check that setting energy_cap_max to 0 also forces this constraint to 0
        m = build_model(
            {
                "techs.test_supply_plus.constraints": {
                    "resource_area_max": 10,
                    "energy_cap_max": 0,
                }
            },
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert (
            m._backend_model.resource_area_constraint["0::test_supply_plus"].upper()
            == 0
        )

    def test_loc_techs_resource_area_per_energy_capacity_constraint(self):
        """
        i for i in sets.loc_techs_area if i in sets.loc_techs_supply_plus
        and constraint_exists(model_run, i, 'constraints.resource_area_per_energy_cap')
        """
        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "resource_area_per_energy_capacity_constraint"
        )

        m = build_model(
            {"techs.test_supply_plus.constraints.resource_area_max": 10},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "resource_area_per_energy_capacity_constraint"
        )

        m = build_model(
            {"techs.test_supply_elec.constraints.resource_area_per_energy_cap": 10},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_area_per_energy_capacity_constraint")

        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "resource_area_per_energy_cap": 10,
                    "resource_area_max": 10,
                }
            },
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_area_per_energy_capacity_constraint")

    def test_locs_resource_area_capacity_per_loc_constraint(self):
        """
        i for i in sets.locs
        if model_run.locations[i].get_key('available_area', None) is not None
        """
        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "resource_area_capacity_per_loc_constraint"
        )

        m = build_model(
            {"locations.0.available_area": 1},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "resource_area_capacity_per_loc_constraint"
        )

        m = build_model(
            {
                "locations.0.available_area": 1,
                "techs.test_supply_plus.constraints.resource_area_max": 10,
            },
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_area_capacity_per_loc_constraint")

    def test_loc_techs_energy_capacity_constraint(self):
        """
        i for i in sets.loc_techs
        if i not in sets.loc_techs_milp + sets.loc_techs_purchase
        """
        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_constraint")

        m2 = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_scale": 5},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m2.run(build_only=True)
        assert (
            m2._backend_model.energy_capacity_constraint["0::test_supply_elec"].upper()
            == m._backend_model.energy_capacity_constraint[
                "0::test_supply_elec"
            ].upper()
            * 5
        )

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_techs_energy_capacity_milp_constraint(self):
        m = build_model(
            {}, "supply_milp,two_hours,investment_costs"
        )  # demand still is in loc_techs
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_constraint")

    def test_loc_techs_energy_capacity_constraint_warning_on_infinite_equals(self):
        # Check that setting `_equals` to infinity is caught:
        override = {
            "locations.0.techs.test_supply_elec.constraints.energy_cap_equals": np.inf
        }
        with pytest.raises(exceptions.ModelError) as error:
            m = build_model(override, "simple_supply,two_hours,investment_costs")
            m.run(build_only=True)

        assert check_error_or_warning(
            error,
            "Cannot use inf for energy_cap_equals for loc:tech `0::test_supply_elec`",
        )

    def test_techs_energy_capacity_systemwide_constraint(self):
        """
        i for i in sets.techs
        if model_run.get_key('techs.{}.constraints.energy_cap_max_systemwide'.format(i), None)
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "energy_capacity_systemwide_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_max_systemwide": 20},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_systemwide_constraint")
        assert (
            "test_supply_elec"
            in m._backend_model.energy_capacity_systemwide_constraint.keys()
        )

        # setting the constraint to infinity leads to Pyomo creating NoConstraint
        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_max_systemwide": np.inf},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_systemwide_constraint")
        assert (
            "test_supply_elec"
            not in m._backend_model.energy_capacity_systemwide_constraint.keys()
        )

        # Check that setting `_equals` to infinity is caught:
        with pytest.raises(exceptions.ModelError) as error:
            m = build_model(
                {
                    "techs.test_supply_elec.constraints.energy_cap_equals_systemwide": np.inf
                },
                "simple_supply,two_hours,investment_costs",
            )
            m.run(build_only=True)

        assert check_error_or_warning(
            error,
            "Cannot use inf for energy_cap_equals_systemwide for tech `test_supply_elec`",
        )

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_equals_systemwide": 20},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_systemwide_constraint")

        # Check that a model without transmission techs doesn't cause an error
        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_equals_systemwide": 20},
            "simple_supply,two_hours,investment_costs",
            model_file="model_minimal.yaml",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_systemwide_constraint")


class TestDispatchConstraints:
    # dispatch.py
    def test_loc_tech_carriers_carrier_production_max_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_production_max_constraint")

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_tech_carriers_carrier_production_max_milp_constraint(self):
        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_max_constraint")

    def test_loc_tech_carriers_carrier_production_min_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_production_min_constraint")

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_tech_carriers_carrier_production_min_milp_constraint(self):
        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "supply_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_constraint")

    def test_loc_tech_carriers_carrier_consumption_max_constraint(self):
        """
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] not in sets.loc_techs_milp
        """

        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_consumption_max_constraint")

    @pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
    def test_loc_tech_carriers_carrier_consumption_max_milp_constraint(self):
        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_consumption_max_constraint")

    def test_loc_techs_resource_max_constraint(self):
        """
        sets.loc_techs_finite_resource_supply_plus,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "resource_max_constraint")

        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_max_constraint")

        m = build_model(
            {"techs.test_supply_plus.constraints.resource": np.inf},
            "simple_supply_and_supply_plus,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "resource_max_constraint")

    def test_loc_techs_storage_max_constraint(self):
        """
        sets.loc_techs_store
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "storage_max_constraint")

        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_max_constraint")

        m = build_model({}, "simple_storage,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_max_constraint")

    def test_loc_tech_carriers_ramping_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i.rsplit('::', 1)[0] in sets.loc_techs_ramping
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "ramping_up_constraint")
        assert not hasattr(m._backend_model, "ramping_down_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_ramping": 0.1},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "ramping_up_constraint")
        assert hasattr(m._backend_model, "ramping_down_constraint")

        m = build_model(
            {"techs.test_conversion.constraints.energy_ramping": 0.1},
            "simple_conversion,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "ramping_up_constraint")
        assert hasattr(m._backend_model, "ramping_down_constraint")


@pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
class TestMILPConstraints:
    # milp.py
    def test_loc_techs_unit_commitment_milp_constraint(self):
        """
        sets.loc_techs_milp,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "unit_commitment_milp_constraint")

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "unit_commitment_milp_constraint")

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "unit_commitment_milp_constraint")

    def test_loc_techs_unit_capacity_milp_constraint(self):
        """
        sets.loc_techs_milp,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "unit_capacity_milp_constraint")

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "unit_capacity_milp_constraint")

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "unit_capacity_milp_constraint")

    def test_loc_tech_carriers_carrier_production_max_milp_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_max_milp_constraint")

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_production_max_milp_constraint")

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_max_milp_constraint")

        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_max_milp_constraint")

    def test_loc_techs_carrier_production_max_conversion_plus_milp_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if i in sets.loc_techs_milp
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_milp_constraint"
        )

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_milp_constraint"
        )

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_milp_constraint"
        )

        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_milp_constraint"
        )

        m = build_model({}, "conversion_plus_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_max_conversion_plus_milp_constraint"
        )

    def test_loc_tech_carriers_carrier_production_min_milp_constraint(self):
        """
        i for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(model_run, i.rsplit('::', 1)[0], 'constraints.energy_cap_min_use')
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
        """
        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_milp_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "supply_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_production_min_milp_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "supply_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_milp_constraint")

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "conversion_plus_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_milp_constraint")

        m = build_model(
            {"techs.test_conversion_plus.constraints.energy_cap_min_use": 0.1},
            "conversion_plus_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_production_min_milp_constraint")

    def test_loc_techs_carrier_production_min_conversion_plus_milp_constraint(self):
        """
        i for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, 'constraints.energy_cap_min_use')
        and i in sets.loc_techs_milp
        """
        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_milp_constraint"
        )

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "supply_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_milp_constraint"
        )

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min_use": 0.1},
            "conversion_plus_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_milp_constraint"
        )

        m = build_model(
            {"techs.test_conversion_plus.constraints.energy_cap_min_use": 0.1},
            "conversion_plus_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_milp_constraint"
        )

        m = build_model(
            {"techs.test_conversion_plus.constraints.energy_cap_min_use": 0.1},
            "conversion_plus_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "carrier_production_min_conversion_plus_milp_constraint"
        )

    def test_loc_tech_carriers_carrier_consumption_max_milp_constraint(self):
        """
        i for i in sets.loc_tech_carriers_con
        if i.rsplit('::', 1)[0] in sets.loc_techs_demand +
            sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit('::', 1)[0] in sets.loc_techs_milp
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_consumption_max_milp_constraint")

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_consumption_max_milp_constraint")

        m = build_model({}, "storage_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "carrier_consumption_max_milp_constraint")

        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "carrier_consumption_max_milp_constraint")

    def test_loc_techs_energy_capacity_units_milp_constraint(self):
        """
        i for i in sets.loc_techs_milp
        if constraint_exists(model_run, i, 'constraints.energy_cap_per_unit')
        is not None
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "energy_capacity_units_milp_constraint")

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_units_milp_constraint")

        m = build_model({}, "storage_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_units_milp_constraint")

        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_units_milp_constraint")

    def test_loc_techs_storage_capacity_units_milp_constraint(self):
        """
        i for i in sets.loc_techs_milp if i in sets.loc_techs_store
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "storage_capacity_units_milp_constraint")

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "storage_capacity_units_milp_constraint")

        m = build_model({}, "storage_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_capacity_units_milp_constraint")

        m = build_model({}, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "storage_capacity_units_milp_constraint")

        m = build_model({}, "supply_and_supply_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "storage_capacity_units_milp_constraint")

    def test_loc_techs_energy_capacity_max_purchase_milp_constraint(self):
        """
        i for i in sets.loc_techs_purchase
        if (constraint_exists(model_run, i, 'constraints.energy_cap_equals') is not None
            or constraint_exists(model_run, i, 'constraints.energy_cap_max') is not None)
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "energy_capacity_max_purchase_milp_constraint"
        )

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "energy_capacity_max_purchase_milp_constraint"
        )

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_max_purchase_milp_constraint")

        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "energy_cap_max": None,
                    "energy_cap_equals": 15,
                }
            },
            "supply_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_max_purchase_milp_constraint")

    def test_loc_techs_energy_capacity_min_purchase_milp_constraint(self):
        """
        i for i in sets.loc_techs_purchase
        if (not constraint_exists(model_run, i, 'constraints.energy_cap_equals')
            and constraint_exists(model_run, i, 'constraints.energy_cap_min'))
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "energy_capacity_min_purchase_milp_constraint"
        )

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "energy_capacity_min_purchase_milp_constraint"
        )

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "energy_capacity_min_purchase_milp_constraint"
        )

        m = build_model(
            {
                "techs.test_supply_elec.constraints": {
                    "energy_cap_max": None,
                    "energy_cap_equals": 15,
                }
            },
            "supply_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "energy_capacity_min_purchase_milp_constraint"
        )

        m = build_model(
            {"techs.test_supply_elec.constraints.energy_cap_min": 10},
            "supply_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "energy_capacity_min_purchase_milp_constraint")

    def test_loc_techs_storage_capacity_max_purchase_milp_constraint(self):
        """
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        """
        m = build_model({}, "simple_storage,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_max_purchase_milp_constraint"
        )

        m = build_model({}, "storage_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_max_purchase_milp_constraint"
        )

        m = build_model({}, "storage_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "storage_capacity_max_purchase_milp_constraint"
        )

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_max_purchase_milp_constraint"
        )

    def test_loc_techs_storage_capacity_min_purchase_milp_constraint(self):
        """
        i for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (not constraint_exists(model_run, i, 'constraints.storage_cap_equals')
            and (constraint_exists(model_run, i, 'constraints.storage_cap_min')
                or constraint_exists(model_run, i, 'constraints.energy_cap_min')))
        """
        m = build_model(
            {"techs.test_storage.constraints.storage_cap_min": 10},
            "simple_storage,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_min_purchase_milp_constraint"
        )

        m = build_model(
            {"techs.test_storage.constraints.storage_cap_min": 10},
            "storage_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_min_purchase_milp_constraint"
        )

        m = build_model({}, "storage_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_min_purchase_milp_constraint"
        )

        m = build_model(
            {"techs.test_storage.constraints.storage_cap_min": 10},
            "storage_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(
            m._backend_model, "storage_capacity_min_purchase_milp_constraint"
        )

        m = build_model(
            {
                "techs.test_storage.constraints": {
                    "storage_cap_equals": 10,
                    "storage_cap_min": 10,
                }
            },
            "storage_purchase,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(
            m._backend_model, "storage_capacity_min_purchase_milp_constraint"
        )

    def test_loc_techs_update_costs_investment_units_milp_constraint(self):
        """
        i for i in sets.loc_techs_milp
        if i in sets.loc_techs_investment_cost and
        any(constraint_exists(model_run, i, 'costs.{}.purchase'.format(j))
               for j in model_run.sets.costs)
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not check_variable_exists(
            m._backend_model, "cost_investment_constraint", "purchased"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_investment_constraint", "units"
        )

        m = build_model({}, "supply_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert not check_variable_exists(
            m._backend_model, "cost_investment_constraint", "purchased"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_investment_constraint", "units"
        )

        m = build_model(
            {"techs.test_supply_elec.costs.monetary.purchase": 1},
            "supply_milp,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not check_variable_exists(
            m._backend_model, "cost_investment_constraint", "purchased"
        )
        assert check_variable_exists(
            m._backend_model, "cost_investment_constraint", "units"
        )

    def test_loc_techs_update_costs_investment_purchase_milp_constraint(self):
        """
        sets.loc_techs_purchase,
        """

        m = build_model({}, "supply_purchase,two_hours,investment_costs")
        m.run(build_only=True)
        assert check_variable_exists(
            m._backend_model, "cost_investment_constraint", "purchased"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_investment_constraint", "units"
        )

    def test_techs_unit_capacity_systemwide_milp_constraint(self):
        """
        sets.techs if unit_cap_max_systemwide or unit_cap_equals_systemwide
        """

        override_max = {
            "links.0,1.exists": True,
            "techs.test_conversion_plus.constraints.units_max_systemwide": 2,
            "locations.1.techs.test_conversion_plus.constraints": {
                "units_max": 2,
                "energy_cap_per_unit": 5,
            },
        }
        override_equals = {
            "links.0,1.exists": True,
            "techs.test_conversion_plus.constraints.units_equals_systemwide": 1,
            "locations.1.techs.test_conversion_plus.costs.monetary.purchase": 1,
        }
        override_equals_inf = {
            "links.0,1.exists": True,
            "techs.test_conversion_plus.constraints.units_equals_systemwide": np.inf,
            "locations.1.techs.test_conversion_plus.costs.monetary.purchase": 1,
        }
        override_transmission = {
            "links.0,1.exists": True,
            "techs.test_transmission_elec.constraints": {
                "units_max_systemwide": 1,
                "lifetime": 25,
            },
            "techs.test_transmission_elec.costs.monetary": {
                "purchase": 1,
                "interest_rate": 0.1,
            },
        }
        override_no_transmission = {
            "techs.test_supply_elec.constraints.units_equals_systemwide": 1,
            "locations.1.techs.test_supply_elec.costs.monetary.purchase": 1,
        }

        m = build_model(override_max, "conversion_plus_milp,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "unit_capacity_systemwide_milp_constraint")
        assert (
            m._backend_model.unit_capacity_systemwide_milp_constraint[
                "test_conversion_plus"
            ].upper()
            == 2
        )

        m = build_model(
            override_equals, "conversion_plus_milp,two_hours,investment_costs"
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "unit_capacity_systemwide_milp_constraint")
        assert (
            m._backend_model.unit_capacity_systemwide_milp_constraint[
                "test_conversion_plus"
            ].lower()
            == 1
        )
        assert (
            m._backend_model.unit_capacity_systemwide_milp_constraint[
                "test_conversion_plus"
            ].upper()
            == 1
        )

        with pytest.raises(ValueError) as error:
            m = build_model(
                override_equals_inf, "conversion_plus_milp,two_hours,investment_costs"
            )
            m.run(build_only=True)
        assert check_error_or_warning(
            error, "Cannot use inf for energy_cap_equals_systemwide"
        )

        m = build_model(
            override_transmission, "simple_supply,two_hours,investment_costs"
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "unit_capacity_systemwide_milp_constraint")
        assert (
            m._backend_model.unit_capacity_systemwide_milp_constraint[
                "test_transmission_elec"
            ].upper()
            == 2
        )

        m = build_model(
            override_no_transmission,
            "simple_supply,two_hours,investment_costs",
            model_file="model_minimal.yaml",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "unit_capacity_systemwide_milp_constraint")

    def test_asynchronous_prod_con_constraint(self):
        """
        Binary switch for prod/con can be activated using the option
        'asynchronous_prod_con'
        """
        m_store = build_model(
            {"techs.test_storage.constraints.force_asynchronous_prod_con": True},
            "simple_storage,investment_costs",
        )
        m_store.run(build_only=True)
        assert hasattr(m_store._backend_model, "prod_con_switch")
        assert hasattr(m_store._backend_model, "asynchronous_con_milp_constraint")
        assert hasattr(m_store._backend_model, "asynchronous_prod_milp_constraint")

        m_trans = build_model(
            {
                "techs.test_transmission_elec.constraints.force_asynchronous_prod_con": True
            },
            "simple_storage,investment_costs",
        )
        m_trans.run(build_only=True)
        assert hasattr(m_trans._backend_model, "prod_con_switch")
        assert hasattr(m_trans._backend_model, "asynchronous_con_milp_constraint")
        assert hasattr(m_trans._backend_model, "asynchronous_prod_milp_constraint")


class TestConversionConstraints:

    # conversion.py
    def test_loc_techs_balance_conversion_constraint(self):
        """
        sets.loc_techs_conversion,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "balance_conversion_constraint")

        m = build_model({}, "simple_conversion,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "balance_conversion_constraint")

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "balance_conversion_constraint")

    def test_loc_techs_cost_var_conversion_constraint(self):
        """
        sets.loc_techs_om_cost_conversion,
        """
        m = build_model(
            {"techs.test_supply_elec.costs.monetary.om_prod": 0.1},
            "simple_supply,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "cost_var_conversion_constraint")

        m = build_model({}, "simple_conversion,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "cost_var_conversion_constraint")

        m = build_model(
            {"techs.test_conversion.costs.monetary.om_prod": 0.1},
            "simple_conversion,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_conversion_constraint")

        assert check_variable_exists(
            m._backend_model, "cost_var_conversion_constraint", "carrier_prod"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_var_conversion_constraint", "carrier_con"
        )

        m = build_model(
            {"techs.test_conversion.costs.monetary.om_con": 0.1},
            "simple_conversion,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "cost_var_conversion_constraint")
        assert check_variable_exists(
            m._backend_model, "cost_var_conversion_constraint", "carrier_con"
        )
        assert not check_variable_exists(
            m._backend_model, "cost_var_conversion_constraint", "carrier_prod"
        )


class TestNetworkConstraints:
    # network.py
    def test_loc_techs_symmetric_transmission_constraint(self):
        """
        sets.loc_techs_transmission,
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run(build_only=True)
        assert hasattr(m._backend_model, "symmetric_transmission_constraint")

        m = build_model({}, "simple_conversion_plus,two_hours,investment_costs")
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "symmetric_transmission_constraint")


class TestPolicyConstraints:
    # policy.py
    def test_techlists_group_share_energy_cap_min_constraint(self):
        """
        i for i in sets.techlists
        if 'energy_cap_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        """
        m = build_model(
            {}, "simple_supply,group_share_energy_cap_min,two_hours,investment_costs"
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "group_share_energy_cap_min_constraint")
        assert not hasattr(m._backend_model, "group_share_energy_cap_max_constraint")
        assert not hasattr(m._backend_model, "group_share_energy_cap_equals_constraint")

    def test_techlists_group_share_energy_cap_max_constraint(self):
        """
        i for i in sets.techlists
        if 'energy_cap_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        """
        m = build_model(
            {}, "simple_supply,group_share_energy_cap_max,two_hours,investment_costs"
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "group_share_energy_cap_min_constraint")
        assert hasattr(m._backend_model, "group_share_energy_cap_max_constraint")
        assert not hasattr(m._backend_model, "group_share_energy_cap_equals_constraint")

    def test_techlists_group_share_energy_cap_equals_constraint(self):
        """
        i for i in sets.techlists
        if 'energy_cap_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        """
        m = build_model(
            {}, "simple_supply,group_share_energy_cap_equals,two_hours,investment_costs"
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "group_share_energy_cap_min_constraint")
        assert not hasattr(m._backend_model, "group_share_energy_cap_max_constraint")
        assert hasattr(m._backend_model, "group_share_energy_cap_equals_constraint")

    def test_techlists_carrier_group_share_carrier_prod_min_constraint(self):
        """
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_min' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_min'.format(i), {}).keys()
        """

        m = build_model(
            {},
            "conversion_and_conversion_plus,group_share_carrier_prod_min,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert hasattr(m._backend_model, "group_share_carrier_prod_min_constraint")
        assert not hasattr(m._backend_model, "group_share_carrier_prod_max_constraint")
        assert not hasattr(
            m._backend_model, "group_share_carrier_prod_equals_constraint"
        )

    def test_techlists_carrier_group_share_carrier_prod_max_constraint(self):
        """
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_max' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_max'.format(i), {}).keys()
        """

        m = build_model(
            {},
            "conversion_and_conversion_plus,group_share_carrier_prod_max,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "group_share_carrier_prod_min_constraint")
        assert hasattr(m._backend_model, "group_share_carrier_prod_max_constraint")
        assert not hasattr(
            m._backend_model, "group_share_carrier_prod_equals_constraint"
        )

    def test_techlists_carrier_group_share_carrier_prod_equals_constraint(self):
        """
        i + '::' + carrier
        for i in sets.techlists
        if 'carrier_prod_equals' in model_run.model.get_key('group_share.{}'.format(i), {}).keys()
        for carrier in sets.carriers
        if carrier in model_run.model.get_key('group_share.{}.carrier_prod_equals'.format(i), {}).keys()
        """

        m = build_model(
            {},
            "conversion_and_conversion_plus,group_share_carrier_prod_equals,two_hours,investment_costs",
        )
        m.run(build_only=True)
        assert not hasattr(m._backend_model, "group_share_carrier_prod_min_constraint")
        assert not hasattr(m._backend_model, "group_share_carrier_prod_max_constraint")
        assert hasattr(m._backend_model, "group_share_carrier_prod_equals_constraint")


# clustering constraints
class TestClusteringConstraints:
    def constraints(self):
        return [
            "balance_storage_inter_cluster_constraint",
            "storage_intra_max_constraint",
            "storage_intra_min_constraint",
            "storage_inter_max_constraint",
            "storage_inter_min_constraint",
        ]

    def decision_variables(self):
        return [
            "storage_inter_cluster",
            "storage_intra_cluster_max",
            "storage_intra_cluster_min",
        ]

    def cluster_model(
        self,
        how="mean",
        storage_inter_cluster=True,
        cyclic=False,
        storage_initial=False,
    ):
        override = {
            "model.subset_time": ["2005-01-01", "2005-01-04"],
            "model.time": {
                "function": "apply_clustering",
                "function_options": {
                    "clustering_func": "file=cluster_days.csv:0",
                    "how": how,
                    "storage_inter_cluster": storage_inter_cluster,
                },
            },
            "run.cyclic_storage": cyclic,
        }
        if storage_initial:
            override.update({"techs.test_storage.constraints.storage_initial": 0})
        return build_model(override, "simple_storage,investment_costs")

    def test_cluster_storage_constraints(self):
        m = self.cluster_model()
        m.run(build_only=True)

        for variable in self.decision_variables():
            assert hasattr(m._backend_model, variable)

        for constraint in self.constraints():
            assert hasattr(m._backend_model, constraint)

        assert not hasattr(m._backend_model, "storage_max_constraint")
        assert not hasattr(m._backend_model, "storage_initial_constraint")

    def test_cluster_cyclic_storage_constraints(self):
        m = self.cluster_model(cyclic=True)
        m.run(build_only=True)

        for variable in self.decision_variables():
            assert hasattr(m._backend_model, variable)

        for constraint in self.constraints():
            assert hasattr(m._backend_model, constraint)

        assert not hasattr(m._backend_model, "storage_max_constraint")
        assert not hasattr(m._backend_model, "storage_initial_constraint")

    def test_no_cluster_storage_constraints(self):
        m = self.cluster_model(storage_inter_cluster=False)
        m.run(build_only=True)

        for variable in self.decision_variables():
            assert not hasattr(m._backend_model, variable)

        for constraint in self.constraints():
            assert not hasattr(m._backend_model, constraint)

        assert hasattr(m._backend_model, "storage_max_constraint")


class TestLogging:
    @pytest.fixture(scope="module")
    def gurobi_model(self):
        pytest.importorskip("gurobipy")
        model_file = os.path.join("model_config_group", "base_model.yaml")
        return build_model(
            model_file=model_file,
            override_dict={"run": {"solver": "gurobi", "solver_io": "python"}},
        )

    def test_no_duplicate_log_message(self, caplog, gurobi_model):
        caplog.set_level(logging.DEBUG)
        gurobi_model.run()
        all_log_messages = [r.msg for r in caplog.records]
        duplicates = [
            item
            for item, count in collections.Counter(all_log_messages).items()
            if count > 1 and item != "" and not item.startswith("Constructing")
        ]
        assert duplicates == []



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

transmission.py
~~~~~~~~~~~~~~~

Plot transmission data.

"""

import pandas as pd
import xarray as xr
import numpy as np

from calliope.preprocess.util import vincenty
from calliope.postprocess.plotting.util import break_name, get_range


def _get_zoom(coordinate_array, width):
    """
    If mapbox is being used for tranmission plotting, get the zoom based on the
    bounding area of the input data and the width (in pixels) of the map
    """

    # Keys are zoom levels, values are m/pixel at that zoom level
    zoom_dict = {
        0: 156412,
        1: 78206,
        2: 39103,
        3: 19551,
        4: 9776,
        5: 4888,
        6: 2444,
        7: 1222,
        8: 610.984,
        9: 305.492,
        10: 152.746,
        11: 76.373,
        12: 38.187,
        13: 19.093,
        14: 9.547,
        15: 4.773,
        16: 2.387,
        17: 1.193,
        18: 0.596,
        19: 0.298,
    }

    bounds = [
        coordinate_array.max(dim="locs").values,
        coordinate_array.min(dim="locs").values,
    ]

    max_distance = vincenty(*bounds)

    metres_per_pixel = max_distance / width

    for k, v in zoom_dict.items():
        if v > metres_per_pixel:
            continue
        else:
            zoom = k - 4
            break

    return zoom


def _get_data(model, var, sum_dims=None):
    var_da = model.get_formatted_array(var).rename({"locs": "locs_to"})

    if sum_dims:
        var_da = var_da.sum(dim=sum_dims)
    techs = list(
        set(model._model_data.techs_transmission.values).intersection(
            var_da.techs.values
        )
    )
    var_df = var_da.loc[dict(techs=techs)].to_pandas()

    clean_var = var_df[
        (var_df != 0)
        & (var_df.columns.isin(model._model_data.techs_transmission.values))
    ]

    clean_var.columns = pd.MultiIndex.from_tuples(
        clean_var.columns.str.split(":").tolist(), names=["techs", "locs_from"]
    )

    return xr.DataArray.from_series(clean_var.stack().stack())


def _fill_scatter(coordinates, energy_cap, scatter_dict, dict_entry, tech):
    mid_edge = (
        lambda _from, _to: (
            coordinates.loc[{"locs": _from, "coordinates": dict_entry}]
            + coordinates.loc[{"locs": _to, "coordinates": dict_entry}]
        ).item()
        / 2
    )

    edge = lambda _from, _to: [
        coordinates.loc[{"locs": _from, "coordinates": dict_entry}].item(),
        coordinates.loc[{"locs": _to, "coordinates": dict_entry}].item(),
        None,
    ]

    links = []
    filled_list = []

    for loc_from in energy_cap.loc[dict(techs=tech)].locs_from:
        for loc_to in energy_cap.loc[dict(techs=tech)].locs_to:
            if [loc_to, loc_from] in links:
                continue
            e_cap = energy_cap.loc[
                dict(techs=tech, locs_to=loc_to, locs_from=loc_from)
            ].fillna(0)
            if e_cap:
                links.append([loc_from, loc_to])
                if dict_entry == "text":
                    e_cap = "inf" if np.isinf(e_cap.item()) else int(e_cap.item())
                    filled_list.append("{} capacity: {}".format(tech, e_cap))
                else:
                    filled_list.append(
                        edge(loc_from, loc_to)
                        if scatter_dict == "edge"
                        else mid_edge(loc_from, loc_to)
                    )

    return filled_list


def _get_centre(coordinates):
    """
    Get centre of a map based on given lat and lon coordinates
    """
    centre = (coordinates.max(dim="locs") + coordinates.min(dim="locs")) / 2

    return dict(
        lat=centre.loc[dict(coordinates="lat")].item(),
        lon=centre.loc[dict(coordinates="lon")].item(),
    )


def plot_transmission(model, mapbox_access_token=None, **kwargs):
    """
    Parameters
    ----------
    mapbox_access_token : str, optional
        If given and a valid Mapbox API key, a Mapbox map is drawn
        for lat-lon coordinates, else (by default), a more simple
        built-in map.

    """
    try:
        coordinates = model._model_data.loc_coordinates.sortby("locs")
    except AttributeError:
        raise ValueError(
            "Model does not define location coordinates "
            "- no transmission plotting possible."
        )

    colors = model._model_data.colors
    names = model._model_data.names

    plot_width = 1000

    if hasattr(model, "results") and len(model.results.data_vars) > 0:
        energy_cap = _get_data(model, "energy_cap")
        carrier_prod = _get_data(
            model, "carrier_prod", sum_dims=["timesteps", "carriers"]
        )
        carrier_con = _get_data(
            model, "carrier_con", sum_dims=["timesteps", "carriers"]
        )
        energy_flow = carrier_con.fillna(0) + carrier_prod.fillna(0)
    else:
        energy_cap = _get_data(model, "energy_cap_max")
        energy_flow = energy_cap.copy()
        energy_flow.loc[dict()] = 0

    if sorted(coordinates.coordinates.values) == ["lat", "lon"]:
        h_coord, v_coord = ("lat", "lon")
        if mapbox_access_token:
            scatter_type = "scattermapbox"
            layout_dict = dict(
                mapbox=dict(
                    accesstoken=mapbox_access_token,
                    center=_get_centre(coordinates),
                    zoom=_get_zoom(coordinates, plot_width),
                    style="light",
                )
            )
        else:
            scatter_type = "scattergeo"
            layout_dict = dict(
                geo=dict(
                    scope="world",
                    projection=dict(type="mercator", scale=1),
                    showland=True,
                    showcountries=True,
                    showsubunits=True,
                    showocean=True,
                    showrivers=True,
                    showlakes=True,
                    lonaxis=dict(range=get_range(coordinates, "lon", 0.1)),
                    lataxis=dict(range=get_range(coordinates, "lat", 0.1)),
                    resolution=50,
                    landcolor="rgba(240, 240, 240, 0.8)",
                    oceancolor="#aec6cf",
                    subunitcolor="blue",
                    countrycolor="green",
                    countrywidth=0.5,
                    subunitwidth=0.5,
                )
            )
    else:
        h_coord, v_coord = ("x", "y")
        scatter_type = "scatter"
        layout_dict = dict()

    mid_edge_scatter_dict = {
        "type": scatter_type,
        "showlegend": False,
        "mode": "markers",
        "hoverinfo": "text",
        "opacity": 0,
    }

    edge_scatter_dict = {
        "type": scatter_type,
        "mode": "lines",
        "hoverinfo": "none",
        "opacity": 0.8,
    }

    data = []

    for tech in sorted(energy_cap.techs.values):
        per_tech_mid_edge_dict = mid_edge_scatter_dict.copy()
        per_tech_mid_edge_dict = {
            **mid_edge_scatter_dict,
            **{
                h_coord: _fill_scatter(
                    coordinates, energy_cap, "mid_edge", h_coord, tech
                ),
                v_coord: _fill_scatter(
                    coordinates, energy_cap, "mid_edge", v_coord, tech
                ),
                "text": _fill_scatter(
                    coordinates, energy_cap, "mid_edge", "text", tech
                ),
                "legendgroup": tech,
                "marker": {"color": colors.loc[dict(techs=tech)].item()},
            },
        }

        h_edge = _fill_scatter(coordinates, energy_cap, "edge", h_coord, tech)
        v_edge = _fill_scatter(coordinates, energy_cap, "edge", v_coord, tech)
        showlegend = True
        for i in range(len(h_edge)):
            data.append(
                {
                    **edge_scatter_dict,
                    **{
                        h_coord: h_edge[i],
                        v_coord: v_edge[i],
                        "showlegend": showlegend,
                        "legendgroup": tech,
                        "name": break_name(names.loc[dict(techs=tech)].item(), 30),
                        "line": {"color": colors.loc[dict(techs=tech)].item()},
                    },
                }
            )
            showlegend = False
        data.append(per_tech_mid_edge_dict)

    node_scatter_dict = {
        h_coord: [
            coordinates.loc[dict(locs=loc, coordinates=h_coord)].item()
            for loc in coordinates.locs
        ],
        v_coord: [
            coordinates.loc[dict(locs=loc, coordinates=v_coord)].item()
            for loc in coordinates.locs
        ],
        "text": [loc.item() for loc in coordinates.locs],
        "name": "Locations",
        "type": scatter_type,
        "legendgroup": "locations",
        "mode": "markers",
        "hoverinfo": "text",
        "marker": {"symbol": "square", "size": 8, "color": "grey"},
    }

    data.append(node_scatter_dict)

    layout_dict.update(
        dict(
            width=plot_width,
            title=model.model_config["name"],
            autosize=True,
            hovermode="closest",
            showlegend=True,
        )
    )

    if kwargs.get("html_only", False):
        del layout_dict["title"]
        del layout_dict["width"]

    return data, layout_dict



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

examples.py
~~~~~~~~~~~

Example models.

"""

import os

from calliope.core.model import Model


_PATHS = {
    "national_scale": os.path.join(
        os.path.dirname(__file__), "example_models", "national_scale"
    ),
    "urban_scale": os.path.join(
        os.path.dirname(__file__), "example_models", "urban_scale"
    ),
}


def national_scale(*args, **kwargs):
    """Returns the built-in national-scale example model."""
    return Model(os.path.join(_PATHS["national_scale"], "model.yaml"), *args, **kwargs)


def time_clustering(*args, **kwargs):
    """Returns the built-in national-scale example model with time clustering."""
    return Model(
        os.path.join(_PATHS["national_scale"], "model.yaml"),
        scenario="time_clustering",
        *args,
        **kwargs,
    )


def time_resampling(*args, **kwargs):
    """Returns the built-in national-scale example model with time resampling."""
    return Model(
        os.path.join(_PATHS["national_scale"], "model.yaml"),
        scenario="time_resampling",
        *args,
        **kwargs,
    )


def urban_scale(*args, **kwargs):
    """Returns the built-in urban-scale example model."""
    return Model(os.path.join(_PATHS["urban_scale"], "model.yaml"), *args, **kwargs)


def milp(*args, **kwargs):
    """Returns the built-in urban-scale example model with MILP constraints enabled."""
    return Model(
        os.path.join(_PATHS["urban_scale"], "model.yaml"),
        scenario="milp",
        *args,
        **kwargs,
    )


def operate(*args, **kwargs):
    """Returns the built-in urban-scale example model in operate mode."""
    return Model(
        os.path.join(_PATHS["urban_scale"], "model.yaml"),
        scenario="operate",
        *args,
        **kwargs,
    )


def time_masking(*args, **kwargs):
    """Returns the built-in urban-scale example model with time masking."""
    return Model(
        os.path.join(_PATHS["urban_scale"], "model.yaml"),
        scenario="time_masking",
        *args,
        **kwargs,
    )



import pytest  # pylint: disable=unused-import

import pyomo.core as po

from calliope.test.common.util import build_test_model as build_model
from calliope.backend.pyomo.util import get_domain, get_param, invalid


@pytest.fixture(scope="class")
def model():
    return build_model({}, "simple_supply,two_hours,investment_costs")


class TestGetParam:
    def test_get_param_with_timestep_existing(self):
        """ """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run()
        param = get_param(
            m._backend_model,
            "resource",
            ("1::test_demand_elec", m._backend_model.timesteps[1]),
        )
        assert po.value(param) == -5  # see demand_elec.csv

    def test_get_param_no_timestep_existing(self):
        """ """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run()
        param = get_param(
            m._backend_model,
            "energy_eff",
            ("1::test_supply_elec", m._backend_model.timesteps[1]),
        )
        assert po.value(param) == 0.9  # see test model.yaml

    def test_get_param_no_timestep_possible(self):
        """ """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run()
        param = get_param(m._backend_model, "energy_cap_max", ("1::test_supply_elec"))
        assert po.value(param) == 10  # see test model.yaml

        param = get_param(
            m._backend_model, "cost_energy_cap", ("monetary", "0::test_supply_elec")
        )
        assert po.value(param) == 10

    def test_get_param_from_default(self):
        """ """
        m = build_model({}, "simple_supply_and_supply_plus,two_hours,investment_costs")
        m.run()

        param = get_param(
            m._backend_model,
            "parasitic_eff",
            ("1::test_supply_plus", m._backend_model.timesteps[1]),
        )
        assert po.value(param) == 1  # see defaults.yaml

        param = get_param(m._backend_model, "resource_cap_min", ("0::test_supply_plus"))
        assert po.value(param) == 0  # see defaults.yaml

        param = get_param(
            m._backend_model, "cost_resource_cap", ("monetary", "1::test_supply_plus")
        )
        assert po.value(param) == 0  # see defaults.yaml

    def test_get_param_no_default_defined(self):
        """
        If a default is not defined, raise KeyError
        """
        m = build_model({}, "simple_supply,two_hours,investment_costs")
        m.run()
        with pytest.raises(KeyError):
            get_param(
                m._backend_model,
                "random_param",
                ("1::test_demand_elec", m._backend_model.timesteps[1]),
            )
            get_param(m._backend_model, "random_param", ("1::test_supply_elec"))


class TestGetDomain:
    @pytest.mark.parametrize(
        "var, domain",
        (
            ("energy_cap_max", "NonNegativeReals"),
            ("resource", "Reals"),
            ("cost_energy_cap", "Reals"),
            ("force_resource", "Boolean"),
            ("names", "Any"),
        ),
    )
    def test_dtypes(self, model, var, domain):
        assert get_domain(model._model_data[var]) == domain


class TestValueValidity:
    def test_invalid(self):
        pyomo_model = po.ConcreteModel()
        pyomo_model.new_set = po.Set(initialize=["a", "b"])
        pyomo_model.new_param = po.Param(
            pyomo_model.new_set,
            initialize={"a": 1},
            mutable=True,
            within=po.NonNegativeReals,
        )

        assert invalid(pyomo_model.new_param["a"]) is False
        assert invalid(pyomo_model.new_param["b"]) is True



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

milp.py
~~~~~~~~~~~~~~~~~

Constraints for binary and integer decision variables

"""

import pyomo.core as po  # pylint: disable=import-error
import numpy as np

from calliope.backend.pyomo.util import (
    get_param,
    get_timestep_weight,
    get_loc_tech,
    split_comma_list,
    loc_tech_is_in,
)

from calliope.backend.pyomo.constraints.capacity import get_capacity_constraint

ORDER = 30  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]
    run_config = backend_model.__calliope_run_config

    if "loc_techs_unit_commitment_milp_constraint" in sets:
        backend_model.unit_commitment_milp_constraint = po.Constraint(
            backend_model.loc_techs_unit_commitment_milp_constraint,
            backend_model.timesteps,
            rule=unit_commitment_milp_constraint_rule,
        )

    if "loc_techs_unit_capacity_milp_constraint" in sets:
        backend_model.unit_capacity_milp_constraint = po.Constraint(
            backend_model.loc_techs_unit_capacity_milp_constraint,
            rule=unit_capacity_milp_constraint_rule,
        )

    if "loc_tech_carriers_carrier_production_max_milp_constraint" in sets:
        backend_model.carrier_production_max_milp_constraint = po.Constraint(
            backend_model.loc_tech_carriers_carrier_production_max_milp_constraint,
            backend_model.timesteps,
            rule=carrier_production_max_milp_constraint_rule,
        )

    if "loc_techs_carrier_production_max_conversion_plus_milp_constraint" in sets:
        backend_model.carrier_production_max_conversion_plus_milp_constraint = po.Constraint(
            backend_model.loc_techs_carrier_production_max_conversion_plus_milp_constraint,
            backend_model.timesteps,
            rule=carrier_production_max_conversion_plus_milp_constraint_rule,
        )

    if "loc_tech_carriers_carrier_consumption_max_milp_constraint" in sets:
        backend_model.carrier_consumption_max_milp_constraint = po.Constraint(
            backend_model.loc_tech_carriers_carrier_consumption_max_milp_constraint,
            backend_model.timesteps,
            rule=carrier_consumption_max_milp_constraint_rule,
        )

    if "loc_tech_carriers_carrier_production_min_milp_constraint" in sets:
        backend_model.carrier_production_min_milp_constraint = po.Constraint(
            backend_model.loc_tech_carriers_carrier_production_min_milp_constraint,
            backend_model.timesteps,
            rule=carrier_production_min_milp_constraint_rule,
        )

    if "loc_techs_carrier_production_min_conversion_plus_milp_constraint" in sets:
        backend_model.carrier_production_min_conversion_plus_milp_constraint = po.Constraint(
            backend_model.loc_techs_carrier_production_min_conversion_plus_milp_constraint,
            backend_model.timesteps,
            rule=carrier_production_min_conversion_plus_milp_constraint_rule,
        )

    if "loc_techs_storage_capacity_units_milp_constraint" in sets:
        backend_model.storage_capacity_units_milp_constraint = po.Constraint(
            backend_model.loc_techs_storage_capacity_units_milp_constraint,
            rule=storage_capacity_units_milp_constraint_rule,
        )

    if "loc_techs_energy_capacity_units_milp_constraint" in sets:
        backend_model.energy_capacity_units_milp_constraint = po.Constraint(
            backend_model.loc_techs_energy_capacity_units_milp_constraint,
            rule=energy_capacity_units_milp_constraint_rule,
        )

    if (
        "loc_techs_update_costs_investment_units_milp_constraint" in sets
        and run_config["mode"] != "operate"
    ):
        for loc_tech, cost in (
            backend_model.loc_techs_update_costs_investment_units_milp_constraint
            * backend_model.costs
        ):

            update_costs_investment_units_milp_constraint(
                backend_model,
                cost,
                loc_tech,
            )

    if (
        "loc_techs_update_costs_investment_purchase_milp_constraint" in sets
        and run_config["mode"] != "operate"
    ):
        for loc_tech, cost in (
            backend_model.loc_techs_update_costs_investment_purchase_milp_constraint
            * backend_model.costs
        ):

            update_costs_investment_purchase_milp_constraint(
                backend_model,
                cost,
                loc_tech,
            )

    if "loc_techs_energy_capacity_max_purchase_milp_constraint" in sets:
        backend_model.energy_capacity_max_purchase_milp_constraint = po.Constraint(
            backend_model.loc_techs_energy_capacity_max_purchase_milp_constraint,
            rule=energy_capacity_max_purchase_milp_constraint_rule,
        )
    if "loc_techs_energy_capacity_min_purchase_milp_constraint" in sets:
        backend_model.energy_capacity_min_purchase_milp_constraint = po.Constraint(
            backend_model.loc_techs_energy_capacity_min_purchase_milp_constraint,
            rule=energy_capacity_min_purchase_milp_constraint_rule,
        )

    if "loc_techs_storage_capacity_max_purchase_milp_constraint" in sets:
        backend_model.storage_capacity_max_purchase_milp_constraint = po.Constraint(
            backend_model.loc_techs_storage_capacity_max_purchase_milp_constraint,
            rule=storage_capacity_max_purchase_milp_constraint_rule,
        )

    if "loc_techs_storage_capacity_min_purchase_milp_constraint" in sets:
        backend_model.storage_capacity_min_purchase_milp_constraint = po.Constraint(
            backend_model.loc_techs_storage_capacity_min_purchase_milp_constraint,
            rule=storage_capacity_min_purchase_milp_constraint_rule,
        )

    if "techs_unit_capacity_systemwide_milp_constraint" in sets:
        backend_model.unit_capacity_systemwide_milp_constraint = po.Constraint(
            backend_model.techs_unit_capacity_systemwide_milp_constraint,
            rule=unit_capacity_systemwide_milp_constraint_rule,
        )

    if "loc_techs_asynchronous_prod_con_milp_constraint" in sets:
        backend_model.asynchronous_con_milp_constraint = po.Constraint(
            backend_model.loc_techs_asynchronous_prod_con_milp_constraint,
            backend_model.timesteps,
            rule=asynchronous_con_milp_constraint_rule,
        )
        backend_model.asynchronous_prod_milp_constraint = po.Constraint(
            backend_model.loc_techs_asynchronous_prod_con_milp_constraint,
            backend_model.timesteps,
            rule=asynchronous_prod_milp_constraint_rule,
        )


def unit_commitment_milp_constraint_rule(backend_model, loc_tech, timestep):
    """
    Constraining the number of integer units
    :math:`operating_units(loc_tech, timestep)` of a technology which
    can operate in a given timestep, based on maximum purchased units
    :math:`units(loc_tech)`

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{operating\\_units}(loc::tech, timestep) \\leq
            \\boldsymbol{units}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{milp},
            \\forall timestep \\in timesteps

    """

    return (
        backend_model.operating_units[loc_tech, timestep]
        <= backend_model.units[loc_tech]
    )


def unit_capacity_milp_constraint_rule(backend_model, loc_tech):
    """
    Add upper and lower bounds for purchased units of a technology

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{units}(loc::tech)
            \\begin{cases}
                = units_{equals}(loc::tech),& \\text{if } units_{equals}(loc::tech)\\\\
                \\leq units_{max}(loc::tech),& \\text{if } units_{max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\quad \\forall loc::tech \\in loc::techs_{milp}

    and (if ``equals`` not enforced):

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{units}(loc::tech) \\geq units_{min}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{milp}

    """
    return get_capacity_constraint(backend_model, "units", loc_tech)


def carrier_production_max_milp_constraint_rule(
    backend_model, loc_tech_carrier, timestep
):
    """
    Set maximum carrier production of MILP techs that aren't conversion plus

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\leq energy_{cap, per unit}(loc::tech) \\times timestep\\_resolution(timestep)
            \\times \\boldsymbol{operating\\_units}(loc::tech, timestep)
            \\times \\eta_{parasitic}(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{milp}, \\forall timestep \\in timesteps

    :math:`\\eta_{parasitic}` is only activated for `supply_plus` technologies
    """
    loc_tech = get_loc_tech(loc_tech_carrier)

    carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    parasitic_eff = get_param(backend_model, "parasitic_eff", (loc_tech, timestep))
    energy_cap = get_param(backend_model, "energy_cap_per_unit", loc_tech)

    return carrier_prod <= (
        backend_model.operating_units[loc_tech, timestep]
        * timestep_resolution
        * energy_cap
        * parasitic_eff
    )


def carrier_production_max_conversion_plus_milp_constraint_rule(
    backend_model, loc_tech, timestep
):
    """
    Set maximum carrier production of conversion_plus MILP techs

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\leq energy_{cap, per unit}(loc::tech) \\times timestep\\_resolution(timestep)
            \\times \\boldsymbol{operating\\_units}(loc::tech, timestep)
            \\times \\eta_{parasitic}(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{milp, conversion^{+}}, \\forall timestep \\in timesteps

    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    energy_cap = get_param(backend_model, "energy_cap_per_unit", loc_tech)
    loc_tech_carriers_out = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"]["out", loc_tech]
    )

    carrier_prod = sum(
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        for loc_tech_carrier in loc_tech_carriers_out
    )

    return carrier_prod <= (
        backend_model.operating_units[loc_tech, timestep]
        * timestep_resolution
        * energy_cap
    )


def carrier_production_min_milp_constraint_rule(
    backend_model, loc_tech_carrier, timestep
):
    """
    Set minimum carrier production of MILP techs that aren't conversion plus

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\geq energy_{cap, per unit}(loc::tech) \\times timestep\\_resolution(timestep)
            \\times \\boldsymbol{operating\\_units}(loc::tech, timestep)
            \\times energy_{cap, min use}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{milp}, \\forall timestep \\in timesteps
    """
    loc_tech = get_loc_tech(loc_tech_carrier)
    carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    min_use = get_param(backend_model, "energy_cap_min_use", (loc_tech, timestep))
    energy_cap = get_param(backend_model, "energy_cap_per_unit", loc_tech)

    return carrier_prod >= (
        backend_model.operating_units[loc_tech, timestep]
        * timestep_resolution
        * energy_cap
        * min_use
    )


def carrier_production_min_conversion_plus_milp_constraint_rule(
    backend_model, loc_tech, timestep
):
    """
    Set minimum carrier production of conversion_plus MILP techs

    .. container:: scrolling-wrapper

        .. math::
            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\geq energy_{cap, per unit}(loc::tech) \\times timestep\\_resolution(timestep)
            \\times \\boldsymbol{operating\\_units}(loc::tech, timestep)
            \\times energy_{cap, min use}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{milp, conversion^{+}},
            \\forall timestep \\in timesteps

    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    energy_cap = get_param(backend_model, "energy_cap_per_unit", loc_tech)
    min_use = get_param(backend_model, "energy_cap_min_use", (loc_tech, timestep))
    loc_tech_carriers_out = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"]["out", loc_tech]
    )

    carrier_prod = sum(
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        for loc_tech_carrier in loc_tech_carriers_out
    )

    return carrier_prod >= (
        backend_model.operating_units[loc_tech, timestep]
        * timestep_resolution
        * energy_cap
        * min_use
    )


def carrier_consumption_max_milp_constraint_rule(
    backend_model, loc_tech_carrier, timestep
):
    """
    Set maximum carrier consumption of demand, storage, and transmission MILP techs

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\geq -1 * energy_{cap, per unit}(loc::tech) \\times timestep\\_resolution(timestep)
            \\times \\boldsymbol{operating\\_units}(loc::tech, timestep)
            \\times \\eta_{parasitic}(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{milp, con}, \\forall timestep \\in timesteps

    """
    loc_tech = get_loc_tech(loc_tech_carrier)
    carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    energy_cap = get_param(backend_model, "energy_cap_per_unit", loc_tech)

    return carrier_con >= (
        -1
        * backend_model.operating_units[loc_tech, timestep]
        * timestep_resolution
        * energy_cap
    )


def energy_capacity_units_milp_constraint_rule(backend_model, loc_tech):
    """
    Set energy capacity decision variable as a function of purchased units

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{energy_{cap}}(loc::tech) =
            \\boldsymbol{units}(loc::tech) \\times energy_{cap, per unit}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{milp}

    """
    return backend_model.energy_cap[loc_tech] == (
        backend_model.units[loc_tech]
        * get_param(backend_model, "energy_cap_per_unit", loc_tech)
    )


def storage_capacity_units_milp_constraint_rule(backend_model, loc_tech):
    """
    Set storage capacity decision variable as a function of purchased units

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{cap}}(loc::tech) =
            \\boldsymbol{units}(loc::tech) \\times storage_{cap, per unit}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{milp, store}

    """
    return backend_model.storage_cap[loc_tech] == (
        backend_model.units[loc_tech]
        * get_param(backend_model, "storage_cap_per_unit", loc_tech)
    )


def energy_capacity_max_purchase_milp_constraint_rule(backend_model, loc_tech):
    """
    Set maximum energy capacity decision variable upper bound as a function of
    binary purchase variable

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\frac{\\boldsymbol{energy_{cap}}(loc::tech)}{energy_{cap, scale}(loc::tech)}
            \\begin{cases}
                = energy_{cap, equals}(loc::tech) \\times \\boldsymbol{purchased}(loc::tech),&
                    \\text{if } energy_{cap, equals}(loc::tech)\\\\
                \\leq energy_{cap, max}(loc::tech) \\times \\boldsymbol{purchased}(loc::tech),&
                    \\text{if } energy_{cap, max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall loc::tech \\in loc::techs_{purchase}

    """
    energy_cap_max = get_param(backend_model, "energy_cap_max", loc_tech)
    energy_cap_equals = get_param(backend_model, "energy_cap_equals", loc_tech)
    energy_cap_scale = get_param(backend_model, "energy_cap_scale", loc_tech)

    if po.value(energy_cap_equals):
        return backend_model.energy_cap[loc_tech] == (
            energy_cap_equals * energy_cap_scale * backend_model.purchased[loc_tech]
        )

    else:
        return backend_model.energy_cap[loc_tech] <= (
            energy_cap_max * energy_cap_scale * backend_model.purchased[loc_tech]
        )


def energy_capacity_min_purchase_milp_constraint_rule(backend_model, loc_tech):
    """
    Set minimum energy capacity decision variable upper bound as a function of
    binary purchase variable

    and (if ``equals`` not enforced):

    .. container:: scrolling-wrapper

        .. math::

            \\frac{\\boldsymbol{energy_{cap}}(loc::tech)}{energy_{cap, scale}(loc::tech)}
            \\geq energy_{cap, min}(loc::tech) \\times \\boldsymbol{purchased}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs
    """
    energy_cap_min = get_param(backend_model, "energy_cap_min", loc_tech)

    energy_cap_scale = get_param(backend_model, "energy_cap_scale", loc_tech)
    return backend_model.energy_cap[loc_tech] >= (
        energy_cap_min * energy_cap_scale * backend_model.purchased[loc_tech]
    )


def storage_capacity_max_purchase_milp_constraint_rule(backend_model, loc_tech):
    """
    Set maximum storage capacity.

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{cap}}(loc::tech)
            \\begin{cases}
                = storage_{cap, equals}(loc::tech) \\times \\boldsymbol{purchased},&
                    \\text{if } storage_{cap, equals} \\\\
                \\leq storage_{cap, max}(loc::tech) \\times \\boldsymbol{purchased},&
                    \\text{if } storage_{cap, max}(loc::tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall loc::tech \\in loc::techs_{purchase, store}

    """
    storage_cap_max = get_param(backend_model, "storage_cap_max", loc_tech)
    storage_cap_equals = get_param(backend_model, "storage_cap_equals", loc_tech)

    if po.value(storage_cap_equals):
        return backend_model.storage_cap[loc_tech] == (
            storage_cap_equals * backend_model.purchased[loc_tech]
        )

    elif po.value(storage_cap_max):
        return backend_model.storage_cap[loc_tech] <= (
            storage_cap_max * backend_model.purchased[loc_tech]
        )

    else:
        return po.Constraint.Skip


def storage_capacity_min_purchase_milp_constraint_rule(backend_model, loc_tech):
    """
    Set minimum storage capacity decision variable as a function of
    binary purchase variable

    if ``equals`` not enforced for storage_cap:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{cap}}(loc::tech)
            \\geq storage_{cap, min}(loc::tech) \\times \\boldsymbol{purchased}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{purchase, store}
    """
    storage_cap_min = get_param(backend_model, "storage_cap_min", loc_tech)

    if po.value(storage_cap_min):
        return backend_model.storage_cap[loc_tech] >= (
            storage_cap_min * backend_model.purchased[loc_tech]
        )

    else:
        return po.Constraint.Skip


def update_costs_investment_units_milp_constraint(backend_model, cost, loc_tech):
    """
    Add MILP investment costs (cost * number of units purchased)

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{investment}}(cost, loc::tech) += \\boldsymbol{units}(loc::tech)
            \\times cost_{purchase}(cost, loc::tech) * timestep_{weight} * depreciation
            \\quad \\forall cost \\in costs, \\forall loc::tech \\in loc::techs_{cost_{investment}, milp}
    """
    model_data_dict = backend_model.__calliope_model_data
    ts_weight = get_timestep_weight(backend_model)
    depreciation_rate = model_data_dict["data"]["cost_depreciation_rate"][
        (cost, loc_tech)
    ]

    cost_purchase = get_param(backend_model, "cost_purchase", (cost, loc_tech))
    cost_of_purchase = (
        backend_model.units[loc_tech] * cost_purchase * ts_weight * depreciation_rate
    )

    if loc_tech_is_in(backend_model, loc_tech, "loc_techs_transmission"):
        cost_of_purchase = cost_of_purchase / 2

    backend_model.cost_investment_rhs[cost, loc_tech].expr += cost_of_purchase

    return None


def update_costs_investment_purchase_milp_constraint(backend_model, cost, loc_tech):
    """
    Add binary investment costs (cost * binary_purchased_unit)

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{investment}}(cost, loc::tech) += \\boldsymbol{purchased}(loc::tech)
            \\times cost_{purchase}(cost, loc::tech) * timestep_{weight} * depreciation
            \\quad \\forall cost \\in costs, \\forall loc::tech \\in loc::techs_{cost_{investment}, purchase}

    """
    model_data_dict = backend_model.__calliope_model_data
    ts_weight = get_timestep_weight(backend_model)
    depreciation_rate = model_data_dict["data"]["cost_depreciation_rate"][
        (cost, loc_tech)
    ]

    cost_purchase = get_param(backend_model, "cost_purchase", (cost, loc_tech))
    cost_of_purchase = (
        backend_model.purchased[loc_tech]
        * cost_purchase
        * ts_weight
        * depreciation_rate
    )

    if loc_tech_is_in(backend_model, loc_tech, "loc_techs_transmission"):
        cost_of_purchase = cost_of_purchase / 2

    backend_model.cost_investment_rhs[cost, loc_tech].expr += cost_of_purchase

    return None


def unit_capacity_systemwide_milp_constraint_rule(backend_model, tech):
    """
    Set constraints to limit the number of purchased units of a single technology
    type across all locations in the model.

    The first valid case is applied:

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc}\\boldsymbol{units}(loc::tech) + \\boldsymbol{purchased}(loc::tech)
            \\begin{cases}
                = units_{equals, systemwide}(tech),&
                    \\text{if } units_{equals, systemwide}(tech)\\\\
                \\leq units_{max, systemwide}(tech),&
                    \\text{if } units_{max, systemwide}(tech)\\\\
                \\text{unconstrained},& \\text{otherwise}
            \\end{cases}
            \\forall tech \\in techs

    """

    if tech in getattr(backend_model, "techs_transmission_names", []):
        all_loc_techs = [
            i
            for i in backend_model.loc_techs_transmission
            if i.split("::")[1].split(":")[0] == tech
        ]
        multiplier = 2  # there are always two technologies associated with one link
    else:
        all_loc_techs = [i for i in backend_model.loc_techs if i.split("::")[1] == tech]
        multiplier = 1

    max_systemwide = get_param(backend_model, "units_max_systemwide", tech)
    equals_systemwide = get_param(backend_model, "units_equals_systemwide", tech)

    if np.isinf(po.value(max_systemwide)) and not equals_systemwide:
        return po.Constraint.NoConstraint
    elif equals_systemwide and np.isinf(po.value(equals_systemwide)):
        raise ValueError(
            "Cannot use inf for energy_cap_equals_systemwide for tech `{}`".format(tech)
        )

    sum_expr_units = sum(
        backend_model.units[loc_tech]
        for loc_tech in all_loc_techs
        if loc_tech_is_in(backend_model, loc_tech, "loc_techs_milp")
    )
    sum_expr_purchase = sum(
        backend_model.purchased[loc_tech]
        for loc_tech in all_loc_techs
        if loc_tech_is_in(backend_model, loc_tech, "loc_techs_purchase")
    )

    if equals_systemwide:
        return sum_expr_units + sum_expr_purchase == equals_systemwide * multiplier
    else:
        return sum_expr_units + sum_expr_purchase <= max_systemwide * multiplier


def asynchronous_con_milp_constraint_rule(backend_model, loc_tech, timestep):
    """
    BigM limit set on `carrier_con`, forcing it to either be zero or non-zero,
    depending on whether `con` is zero or one, respectively.

    .. container:: scrolling-wrapper

        .. math::
            - \\boldsymbol{carrier_con}[loc::tech::carrier, timestep] \\leq
            \\text{bigM} \\times (1 - \\boldsymbol{prod_con_switch}[loc::tech, timestep])
            \\forall loc::tech \\in loc::techs_{asynchronous_prod_con},
            \\forall timestep \\in timesteps

    """
    model_dict = backend_model.__calliope_model_data
    loc_tech_carrier = model_dict["data"]["lookup_loc_techs"][loc_tech]

    return (
        -1 * backend_model.carrier_con[loc_tech_carrier, timestep]
        <= (1 - backend_model.prod_con_switch[loc_tech, timestep]) * backend_model.bigM
    )


def asynchronous_prod_milp_constraint_rule(backend_model, loc_tech, timestep):
    """
    BigM limit set on `carrier_prod`, forcing it to either be zero or non-zero,
    depending on whether `prod` is zero or one, respectively.

    .. container:: scrolling-wrapper

        .. math::
            \\boldsymbol{carrier_prod}[loc::tech::carrier, timestep] \\leq
            \\text{bigM} \\times \\boldsymbol{prod_con_switch}[loc::tech, timestep]
            \\forall loc::tech \\in loc::techs_{asynchronous_prod_con},
            \\forall timestep \\in timesteps

    """
    model_dict = backend_model.__calliope_model_data
    loc_tech_carrier = model_dict["data"]["lookup_loc_techs"][loc_tech]

    return (
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        <= backend_model.prod_con_switch[loc_tech, timestep] * backend_model.bigM
    )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

import pyomo.core as po  # pylint: disable=import-error


def initialize_decision_variables(backend_model):
    """
    Defines decision variables.

    ==================== ========================================
    Variable             Dimensions
    ==================== ========================================
    energy_cap           loc_techs
    carrier_prod         loc_tech_carriers_prod, timesteps
    carrier_con          loc_tech_carriers_con, timesteps
    cost                 costs, loc_techs_cost
    resource_area        loc_techs_area,
    storage_cap          loc_techs_store
    storage              loc_techs_store, timesteps
    resource_con         loc_techs_supply_plus, timesteps
    resource_cap         loc_techs_supply_plus
    carrier_export       loc_tech_carriers_export, timesteps
    cost_var             costs, loc_techs_om_cost, timesteps
    cost_investment      costs, loc_techs_investment_cost
    purchased            loc_techs_purchase
    units                loc_techs_milp
    operating\\_units     loc_techs_milp, timesteps
    unmet\\_demand        loc_carriers, timesteps
    unused\\_supply       loc_carriers, timesteps
    ==================== ========================================

    """
    model_data_dict = backend_model.__calliope_model_data
    run_config = backend_model.__calliope_run_config

    ##
    # Variables which are always assigned
    ##
    if run_config["mode"] != "operate":
        backend_model.energy_cap = po.Var(
            backend_model.loc_techs, within=po.NonNegativeReals
        )
    backend_model.carrier_prod = po.Var(
        backend_model.loc_tech_carriers_prod,
        backend_model.timesteps,
        within=po.NonNegativeReals,
    )
    backend_model.carrier_con = po.Var(
        backend_model.loc_tech_carriers_con,
        backend_model.timesteps,
        within=po.NegativeReals,
    )
    backend_model.cost = po.Var(
        backend_model.costs, backend_model.loc_techs_cost, within=po.Reals
    )

    ##
    # Conditionally assigned variables
    ##

    if "loc_techs_area" in model_data_dict["sets"] and run_config["mode"] != "operate":
        backend_model.resource_area = po.Var(
            backend_model.loc_techs_area, within=po.NonNegativeReals
        )

    if "loc_techs_store" in model_data_dict["sets"]:
        if run_config["mode"] != "operate":
            backend_model.storage_cap = po.Var(
                backend_model.loc_techs_store, within=po.NonNegativeReals
            )
        if hasattr(backend_model, "clusters") and hasattr(backend_model, "datesteps"):
            backend_model.storage_inter_cluster = po.Var(
                backend_model.loc_techs_store,
                backend_model.datesteps,
                within=po.NonNegativeReals,
            )
            backend_model.storage_intra_cluster_max = po.Var(
                backend_model.loc_techs_store, backend_model.clusters, within=po.Reals
            )
            backend_model.storage_intra_cluster_min = po.Var(
                backend_model.loc_techs_store, backend_model.clusters, within=po.Reals
            )
            storage_within = po.Reals
        else:
            storage_within = po.NonNegativeReals
        backend_model.storage = po.Var(
            backend_model.loc_techs_store,
            backend_model.timesteps,
            within=storage_within,
        )

    if "loc_techs_supply_plus" in model_data_dict["sets"]:
        backend_model.resource_con = po.Var(
            backend_model.loc_techs_supply_plus,
            backend_model.timesteps,
            within=po.NonNegativeReals,
        )
        if run_config["mode"] != "operate":
            backend_model.resource_cap = po.Var(
                backend_model.loc_techs_supply_plus, within=po.NonNegativeReals
            )

    if "loc_techs_export" in model_data_dict["sets"]:
        backend_model.carrier_export = po.Var(
            backend_model.loc_tech_carriers_export,
            backend_model.timesteps,
            within=po.NonNegativeReals,
        )

    if "loc_techs_om_cost" in model_data_dict["sets"]:
        backend_model.cost_var = po.Var(
            backend_model.costs,
            backend_model.loc_techs_om_cost,
            backend_model.timesteps,
            within=po.Reals,
        )

    if (
        "loc_techs_investment_cost" in model_data_dict["sets"]
        and run_config["mode"] != "operate"
    ):
        backend_model.cost_investment = po.Var(
            backend_model.costs,
            backend_model.loc_techs_investment_cost,
            within=po.Reals,
        )

    if (
        "loc_techs_purchase" in model_data_dict["sets"]
        and run_config["mode"] != "operate"
    ):
        backend_model.purchased = po.Var(
            backend_model.loc_techs_purchase, within=po.Binary
        )

    if (
        "group_demand_share_per_timestep_decision" in model_data_dict["data"]
        and run_config["mode"] != "operate"
    ):
        backend_model.demand_share_per_timestep_decision = po.Var(
            backend_model.loc_tech_carriers_prod, within=po.NonNegativeReals
        )

    if "loc_techs_milp" in model_data_dict["sets"]:
        if run_config["mode"] != "operate":
            backend_model.units = po.Var(
                backend_model.loc_techs_milp, within=po.NonNegativeIntegers
            )
        backend_model.operating_units = po.Var(
            backend_model.loc_techs_milp,
            backend_model.timesteps,
            within=po.NonNegativeIntegers,
        )
        # For any milp tech, we need to update energy_cap, as energy_cap_max and energy_cap_equals
        # are replaced by energy_cap_per_unit
        if run_config["mode"] == "operate":
            for k, v in backend_model.units.items():
                backend_model.energy_cap[k] = v * backend_model.energy_cap_per_unit[k]

    if "loc_techs_asynchronous_prod_con" in model_data_dict["sets"]:
        backend_model.prod_con_switch = po.Var(
            backend_model.loc_techs_asynchronous_prod_con,
            backend_model.timesteps,
            within=po.Binary,
        )
        backend_model.bigM = run_config.get("bigM", 1e10)

    if run_config.get("ensure_feasibility", False):
        backend_model.unmet_demand = po.Var(
            backend_model.loc_carriers,
            backend_model.timesteps,
            within=po.NonNegativeReals,
        )
        backend_model.unused_supply = po.Var(
            backend_model.loc_carriers, backend_model.timesteps, within=po.NegativeReals
        )
        backend_model.bigM = run_config.get("bigM", 1e10)
